#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: test_augers.py
@time: 2020/2/5 2:24
@desc:
"""
import pytest

import numpy as np
from colorama import Fore, Style
import matplotlib.pyplot as plt
import cv2

from alchemy_cat.contrib.voc import VOCAug, label_map2color_map
from alchemy_cat.data import DataAuger
from alchemy_cat.data.plugins.augers import RandMirror, MultiMirror, RandColorJitter, RandUpDown, MultiUpDown, \
    int_img2float32_img, centralize, pad_img_label, RandCrop, FiveCrop, RandScale, MultiScale
from alchemy_cat.dag import Node
from alchemy_cat.acplot.figure_wall import RowFigureWall, ColumnFigureWall
from alchemy_cat.py_tools import set_numpy_rand_seed
from alchemy_cat.alg import find_nearest_odd_size, find_nearest_even_size


def plot_img_label_pair(img_pair, label_pair=None, img_num=None):
    img_wall = RowFigureWall(img_pair, space_width=20, pad_location='center', color_channel_order='BGR')
    label_wall = RowFigureWall(label_pair, space_width=20, pad_location='center', color_channel_order='RGB') \
        if label_pair is not None else None

    show_wall = ColumnFigureWall([img_wall, label_wall], space_width=20, color_channel_order='BGR') \
        if label_wall is not None else img_wall
    show_wall.plot(num=img_num)
    plt.show()


class BaseAuger(DataAuger):
    def build_graph(self):
        @self.graph.register(inputs=['example'], outputs=['img_id', 'img', 'label'])
        def split_example(example):
            return example


@pytest.fixture(scope="module")
def voc_dataset():
    return VOCAug(split='val')


def setup_function(func):
    set_numpy_rand_seed(0)
    print(f"-----------setup function: set_numpy_rand_seed(0)-----------")


def test_rand_mirror(voc_dataset):
    """Test rand mirror with following steps:
        1. Create a rand_mirror auger on voc val dataset
        2. Test is all img is mirrored accroding to it's rand_seed
        3. Test is rand distribution is random and balanced
        4. Visualization

    Args:
        voc_dataset: VOCAug val dataset
    """
    print(Fore.LIGHTYELLOW_EX + "===========Test Rand Mirror==========" + Style.RESET_ALL)

    class RandMirrorAuger(BaseAuger):

        def build_graph(self):
            super(RandMirrorAuger, self).build_graph()

            self.graph.add_node(RandMirror, inputs=['img', 'label'], outputs=['mirrored_img', 'mirrored_label'])

    auger = RandMirrorAuger(voc_dataset, verbosity=0, slim=True)
    mirror_node: Node = auger.rand_nodes[0]

    mirrored, not_mirrored = 0, 0
    for index, outputs in enumerate(auger):
        mirrored_img, mirrored_label = outputs
        _, img, label = voc_dataset[index]

        rand_seed = auger.rand_seeds[mirror_node.id]

        assert np.all(mirrored_img[:, ::rand_seed] == img), f"Current index={index}, rand_seed={rand_seed}"
        assert np.all(mirrored_label[:, ::rand_seed] == label), f"Current index={index}, rand_seed={rand_seed}"

        if index % 100 == 0:
            print(Fore.BLUE + f"{index // 100}: rand_seed = {rand_seed}\n" + Style.RESET_ALL)
            img_pair = [img, mirrored_img]
            label_pair = map(label_map2color_map, [label, mirrored_label])

            plot_img_label_pair(img_pair, label_pair, img_num=f"ID: {_}; index: {index}")

        if rand_seed == 1:
            not_mirrored += 1
        elif rand_seed == -1:
            mirrored += 1
        else:
            raise ValueError(f"rand_seed={rand_seed} is not 1 or -1")

    assert mirrored + not_mirrored == len(voc_dataset)
    assert mirrored / len(voc_dataset) == pytest.approx(0.5, abs=0.04)
    print(Fore.LIGHTYELLOW_EX + f"mirrored: {mirrored}; not_mirrored: {not_mirrored}" + Style.RESET_ALL)


def test_weighted_rand_mirror(voc_dataset):
    """Test weighted rand mirror with following steps:
        1. Create a rand_mirror auger on voc val dataset
        2. Test is all img is mirrored accroding to it's rand_seed
        3. Test is rand distribution is random and weighted
        4. Visualization

    Args:
        voc_dataset: VOCAug val dataset
    """
    print(Fore.LIGHTYELLOW_EX + "===========Test Weighted Rand Mirror==========" + Style.RESET_ALL)

    class WeightedRandMirror(RandMirror):
        weight = [90, 10]

    class WeightedRandMirrorAuger(BaseAuger):

        def build_graph(self):
            super(WeightedRandMirrorAuger, self).build_graph()

            self.graph.add_node(WeightedRandMirror,
                                inputs=['img', 'label'], outputs=['mirrored_img', 'mirrored_label'])

    auger = WeightedRandMirrorAuger(voc_dataset, verbosity=0, slim=True)
    mirror_node: Node = auger.rand_nodes[0]

    mirrored, not_mirrored = 0, 0
    for index, outputs in enumerate(auger):
        mirrored_img, mirrored_label = outputs
        _, img, label = voc_dataset[index]

        rand_seed = auger.rand_seeds[mirror_node.id]

        assert np.all(mirrored_img[:, ::rand_seed] == img), f"Current index={index}, rand_seed={rand_seed}"
        assert np.all(mirrored_label[:, ::rand_seed] == label), f"Current index={index}, rand_seed={rand_seed}"

        if index % 100 == 0:
            print(Fore.BLUE + f"{index // 100}: rand_seed = {rand_seed}\n" + Style.RESET_ALL)
            img_pair = [img, mirrored_img]
            label_pair = map(label_map2color_map, [label, mirrored_label])

            plot_img_label_pair(img_pair, label_pair, img_num=f"ID: {_}; index: {index}")

        if rand_seed == 1:
            not_mirrored += 1
        elif rand_seed == -1:
            mirrored += 1
        else:
            raise ValueError(f"rand_seed={rand_seed} is not 1 or -1")

    assert mirrored + not_mirrored == len(voc_dataset)
    assert not_mirrored / len(voc_dataset) == pytest.approx(0.9, abs=0.04)
    print(Fore.LIGHTYELLOW_EX + f"mirrored: {mirrored}; not_mirrored: {not_mirrored}" + Style.RESET_ALL)


def test_rand_log(voc_dataset):
    """Test rand mirror with rand log following steps:
        1. Create two rand_mirror auger on voc val dataset
        2. Step0: Test is all img is mirrored accroding to it's rand_seed
        3. Step0: Test is rand distribution is random and balanced
        4. Step1: Test is all img is mirrored accroding to it's rand_seed
        5. Step1: Test is rand distribution is random and balanced
        6. Test is Step0's random is the same to step1's random
        7. Visualization

    Args:
        voc_dataset: VOCAug val dataset
    """
    print(Fore.LIGHTYELLOW_EX + "===========Test Rand Log==========" + Style.RESET_ALL)

    class RandMirrorAuger(BaseAuger):

        def build_graph(self):
            super(RandMirrorAuger, self).build_graph()

            self.graph.add_node(RandMirror, inputs=['img', 'label'], outputs=['mirrored_img', 'mirrored_label'])

    auger0 = RandMirrorAuger(voc_dataset, verbosity=0, slim=True, rand_seed_log='./Temp/rand_log')
    mirror_node0: Node = auger0.rand_nodes[0]

    rand_seed_record0 = []
    for index, outputs in enumerate(auger0):
        mirrored_img, mirrored_label = outputs
        _, img, label = voc_dataset[index]

        rand_seed = auger0.rand_seeds[mirror_node0.id]

        assert np.all(mirrored_img[:, ::rand_seed] == img), f"Current index={index}, rand_seed={rand_seed}"
        assert np.all(mirrored_label[:, ::rand_seed] == label), f"Current index={index}, rand_seed={rand_seed}"

        if index % 300 == 0:
            print(Fore.BLUE + f"{index // 100}: rand_seed = {rand_seed}\n" + Style.RESET_ALL)
            img_pair = [img, mirrored_img]
            label_pair = map(label_map2color_map, [label, mirrored_label])

            plot_img_label_pair(img_pair, label_pair, img_num=f"ID: {_}; index: {index}")

        rand_seed_record0.append(rand_seed)

    assert len(rand_seed_record0) == len(voc_dataset)
    assert rand_seed_record0.count(1) / len(voc_dataset) == pytest.approx(0.5, abs=0.04)

    auger1 = RandMirrorAuger(voc_dataset, verbosity=0, slim=True, rand_seed_log='./Temp/rand_log')
    mirror_node1: Node = auger1.rand_nodes[0]

    rand_seed_record1 = []
    for index, outputs in enumerate(auger1):
        mirrored_img, mirrored_label = outputs
        _, img, label = voc_dataset[index]

        rand_seed = auger1.rand_seeds[mirror_node1.id]

        assert np.all(mirrored_img[:, ::rand_seed] == img), f"Current index={index}, rand_seed={rand_seed}"
        assert np.all(mirrored_label[:, ::rand_seed] == label), f"Current index={index}, rand_seed={rand_seed}"

        if index % 300 == 0:
            print(Fore.BLUE + f"{index // 100}: rand_seed = {rand_seed}\n" + Style.RESET_ALL)
            img_pair = [img, mirrored_img]
            label_pair = map(label_map2color_map, [label, mirrored_label])

            plot_img_label_pair(img_pair, label_pair, img_num=f"ID: {_}; index: {index}")

        rand_seed_record1.append(rand_seed)

    assert len(rand_seed_record1) == len(voc_dataset)
    assert rand_seed_record1.count(1) / len(voc_dataset) == pytest.approx(0.5, abs=0.04)

    assert rand_seed_record0 == rand_seed_record1


def test_multi_mirror(voc_dataset):
    """Test multi mirror with following steps:
        1. Create a multi mirror auger on voc test dataset
        2. Test whether all img is mirrored accroding to it's output_index
        3. Test whether output_index is met with dataset index and auger index
        4. Test whether the total number is right
        5. Visualization

    Args:
        voc_dataset: VOCAug val dataset
    """
    print(Fore.LIGHTYELLOW_EX + "===========Test Multi Mirror==========" + Style.RESET_ALL)

    class MultiMirrorAuger(BaseAuger):

        def build_graph(self):
            super(MultiMirrorAuger, self).build_graph()

            self.graph.add_node(MultiMirror, inputs=['img', 'label'], outputs=['mirrored_img', 'mirrored_label'])

    auger = MultiMirrorAuger(voc_dataset, verbosity=0, slim=True)
    multi_node: Node = auger.multi_nodes[0]

    mirrored, not_mirrored = 0, 0
    for auger_index, outputs in enumerate(auger):
        mirrored_img, mirrored_label = outputs

        dataset_index = auger_index // 2
        _, img, label = voc_dataset[dataset_index]

        output_index = multi_node._fct.output_index
        assert output_index == auger_index - 2 * dataset_index

        mirror = 1 if output_index == 0 else -1
        assert np.all(mirrored_img[:, ::mirror] == img), f"Current index={dataset_index}, mirror={mirror}"
        assert np.all(mirrored_label[:, ::mirror] == label), f"Current dataset_index={dataset_index}, mirror={mirror}"

        if dataset_index % 99 == 0:
            print(Fore.BLUE + f"{dataset_index // 99}: mirror = {mirror}\n" + Style.RESET_ALL)
            img_pair = [img, mirrored_img]
            label_pair = map(label_map2color_map, [label, mirrored_label])

            plot_img_label_pair(img_pair, label_pair)

        if mirror == 1:
            not_mirrored += 1
        elif mirror == -1:
            mirrored += 1
        else:
            raise ValueError(f"mirror={mirror} is not 1 or -1")

    assert mirrored + not_mirrored == 2 * len(voc_dataset)
    assert mirrored == not_mirrored
    print(Fore.LIGHTYELLOW_EX + f"mirrored: {mirrored}; not_mirrored: {not_mirrored}" + Style.RESET_ALL)


def test_rand_updown(voc_dataset):
    """Test rand mirror with following steps:
        1. Create a rand updown auger on voc val dataset
        2. Test is all img is updown accroding to it's rand_seed
        3. Test is rand distribution is random and balanced
        4. Visualization

    Args:
        voc_dataset: VOCAug val dataset
    """
    print(Fore.LIGHTYELLOW_EX + "===========Test Rand UpDown==========" + Style.RESET_ALL)

    class RandUpDownAuger(BaseAuger):

        def build_graph(self):
            super(RandUpDownAuger, self).build_graph()

            self.graph.add_node(RandUpDown, inputs=['img', 'label'], outputs=['updown_img', 'updown_label'])

    auger = RandUpDownAuger(voc_dataset, verbosity=0, slim=True)
    updown_node: Node = auger.rand_nodes[0]

    updown, not_updown = 0, 0
    for index, outputs in enumerate(auger):
        updown_img, updown_label = outputs
        _, img, label = voc_dataset[index]

        rand_seed = auger.rand_seeds[updown_node.id]

        assert np.all(updown_img[::rand_seed] == img), f"Current index={index}, rand_seed={rand_seed}"
        assert np.all(updown_label[::rand_seed] == label), f"Current index={index}, rand_seed={rand_seed}"

        if index % 100 == 0:
            print(Fore.BLUE + f"{index // 100}: rand_seed = {rand_seed}\n" + Style.RESET_ALL)
            img_pair = [img, updown_img]
            label_pair = map(label_map2color_map, [label, updown_label])

            plot_img_label_pair(img_pair, label_pair)

        if rand_seed == 1:
            not_updown += 1
        elif rand_seed == -1:
            updown += 1
        else:
            raise ValueError(f"rand_seed={rand_seed} is not 1 or -1")

    assert updown + not_updown == len(voc_dataset)
    assert updown / len(voc_dataset) == pytest.approx(0.5, abs=0.04)
    print(Fore.LIGHTYELLOW_EX + f"updown: {updown}; not_updown: {not_updown}" + Style.RESET_ALL)


def test_multi_updown(voc_dataset):
    """Test multi mirror with following steps:
        1. Create a multi updown auger on voc test dataset
        2. Test whether all img is updown accroding to it's output_index
        3. Test whether output_index is met with dataset index and auger index
        4. Test whether the total number is right
        5. Visualization

    Args:
        voc_dataset: VOCAug val dataset
    """
    print(Fore.LIGHTYELLOW_EX + "===========Test Multi UpDown==========" + Style.RESET_ALL)

    class MultiUpDownAuger(BaseAuger):

        def build_graph(self):
            super(MultiUpDownAuger, self).build_graph()

            self.graph.add_node(MultiUpDown, inputs=['img', 'label'], outputs=['updown_img', 'updown_label'])

    auger = MultiUpDownAuger(voc_dataset, verbosity=0, slim=True)
    multi_node: Node = auger.multi_nodes[0]

    updown_count, not_updown_count = 0, 0
    for auger_index, outputs in enumerate(auger):
        updown_img, updown_label = outputs

        dataset_index = auger_index // 2
        _, img, label = voc_dataset[dataset_index]

        output_index = multi_node._fct.output_index
        assert output_index == auger_index - 2 * dataset_index

        updown = 1 if output_index == 0 else -1
        assert np.all(updown_img[::updown, ...] == img), f"Current index={dataset_index}, updown={updown}"
        assert np.all(updown_label[::updown, ...] == label), f"Current dataset_index={dataset_index}, updown={updown}"

        if dataset_index % 99 == 0:
            print(Fore.BLUE + f"{dataset_index // 99}: updown = {updown}\n" + Style.RESET_ALL)
            img_pair = [img, updown_img]
            label_pair = map(label_map2color_map, [label, updown_label])

            plot_img_label_pair(img_pair, label_pair)

        if updown == 1:
            not_updown_count += 1
        elif updown == -1:
            updown_count += 1
        else:
            raise ValueError(f"mirror={updown} is not 1 or -1")

    assert updown_count + not_updown_count == 2 * len(voc_dataset)
    assert updown_count == not_updown_count
    print(Fore.LIGHTYELLOW_EX + f"updown: {updown_count}; not_updown: {not_updown_count}" + Style.RESET_ALL)


def test_rand_color_jitter(voc_dataset):
    """Test RandColorJitter
        1. Create RandColorJitter Auger
        2. Reverse the jitter process accroding to rand seed, then compare to the origin img.
            - Except ignore pixel, warning if approx equal (abs=5) pixel ratio (also called reversible ratio) > 0.8
            - Except ignore pixel, macro mean reversible ratio > 0.8
        3. Test the whether the distribution is random and balanced
        4. Make sure the distortion ration is not too high under default parameter
            - Warning if distortion ratio >= 0.7 for single img.
            - Macro mean distortion ratio < 0.2
        5. Total output number should be correct
        6. Visualization

    Args:
        voc_dataset: VOCAug val dataset
    """
    print(Fore.LIGHTYELLOW_EX + "===========Test Color Jitter==========" + Style.RESET_ALL)

    class RandColorJitterAuger(BaseAuger):

        def build_graph(self):
            super(RandColorJitterAuger, self).build_graph()

            self.graph.add_node(RandColorJitter, inputs=['img'], outputs=['jitter_img'])

    auger = RandColorJitterAuger(voc_dataset, verbosity=0, slim=True)
    jitter_node: Node = auger.rand_nodes[0]

    def dejitter_bright(delta_bright, img, ignore):
        if delta_bright > 0:
            ignore |= np.any((img == 255), axis=2)
        else:
            ignore |= np.any((img == 0), axis=2)
        return cv2.convertScaleAbs(img, alpha=1, beta=(- delta_bright))

    def dejitter_contract(mul_contract, img, ignore):
        if mul_contract > 1:
            ignore |= np.any((img == 255), axis=2)
        return cv2.convertScaleAbs(img, alpha=1 / mul_contract, beta=0)

    def dejitter_saturate(mul_saturate, img, ignore):
        img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
        if mul_saturate > 1:
            ignore |= np.any((img == 255), axis=2)
        img[:, :, 1] = cv2.convertScaleAbs(img[:, :, 1], alpha=(1 / mul_saturate), beta=0)
        return cv2.cvtColor(img, cv2.COLOR_HSV2BGR)

    def dejitter_hue(delta_hue, img, ignore):
        img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV_FULL)
        if delta_hue > 0:
            ignore |= np.any((img == 255), axis=2)
        if delta_hue < 0:
            ignore |= np.any((img == 0), axis=2)
        img[:, :, 0] = cv2.convertScaleAbs(img[:, :, 0], alpha=1, beta=(- delta_hue))
        return cv2.cvtColor(img, cv2.COLOR_HSV2BGR_FULL)

    sum_order = np.zeros(4, dtype=np.int64)
    sum_activated = np.zeros(4, dtype=np.int64)

    distortion_ratios = []
    reversible_ratios = []
    for index, jitter_img in enumerate(auger):
        _, img, _ = voc_dataset[index]

        rand_seed = auger.rand_seeds[jitter_node.id]
        jitter_order = rand_seed['jitter_order']
        sum_order += np.array(jitter_order, dtype=np.int64)

        # * Dejitter
        distortion = np.zeros(img.shape[:2], dtype=np.bool_)
        dejitter_img = jitter_img
        delta_bright, mul_contract, mul_saturate, delta_hue = \
            map(lambda k: rand_seed.get(k), ['delta_bright', 'mul_contract', 'mul_saturate', 'delta_hue'])

        for order in jitter_order[::-1]:
            if order == 0 and delta_bright is not None:
                sum_activated[order] += 1
                dejitter_img = dejitter_bright(delta_bright, dejitter_img, distortion)
            elif order == 1 and mul_contract is not None:
                sum_activated[order] += 1
                dejitter_img = dejitter_contract(mul_contract, dejitter_img, distortion)
            elif order == 2 and mul_saturate is not None:
                sum_activated[order] += 1
                dejitter_img = dejitter_saturate(mul_saturate, dejitter_img, distortion)
            elif order == 3 and delta_hue is not None:
                sum_activated[order] += 1
                dejitter_img = dejitter_hue(delta_hue, dejitter_img, distortion)

        # * Visualization
        def vis(problem):
            print(Fore.LIGHTYELLOW_EX + f"{index} [{problem}]: rand_seed = {rand_seed}" + Style.RESET_ALL)
            plot_img_label_pair([img, jitter_img, dejitter_img])

        # * Test distortion ratio
        distortion_ratio = np.sum(distortion) / distortion.size
        if distortion_ratio >= 0.7:
            vis(f"distortion_ratio = {distortion_ratio} >= 0.7")
        # assert distortion_ratio < 0.7
        distortion_ratios.append(distortion_ratio)

        # * Test reversible ratio
        reversible_without_distortion = np.all(np.isclose(dejitter_img, img, atol=5.0, rtol=0), axis=2) & (~distortion)
        reversible_ratio = np.sum(reversible_without_distortion) / (distortion.size - np.sum(distortion))
        if reversible_ratio <= 0.65:
            vis(f"reversible_ratio = {reversible_ratio} <= 0.65")
        # assert reversible_ratio > 0.65
        reversible_ratios.append(reversible_ratio)

    # * Test total num
    assert len(distortion_ratios) == len(voc_dataset)

    # * Test whether distribution is random and balanced
    assert sum_activated / len(voc_dataset) == pytest.approx(0.5, abs=0.04)
    assert sum_order / len(voc_dataset) == pytest.approx(1.5, abs=0.04)

    # * Test macro mean distortion and reversible ratio
    assert np.mean(distortion_ratios) < 0.2
    assert np.nanmean(reversible_ratios) > 0.8


def test_int_img2float32_img(voc_dataset):
    """Test int img converting to float32 img
        1. Test output img is float32
        2. Test output img has the same value as int img

    Args:
        voc_dataset: VOC Aug val dataset
    """
    _, img, _ = voc_dataset[0]

    float_img = int_img2float32_img(img)
    assert float_img.dtype == np.float32
    assert np.all(float_img.astype(np.uint8) == img)


def test_centralize(voc_dataset):
    """Test int img converting to float32 img
        1. centralized img could be decentralized correctly accroding to mean and std

    Args:
        voc_dataset: VOC Aug val dataset
    """
    _, img, _ = voc_dataset[0]

    float_img = int_img2float32_img(img)
    centralized_img = centralize(float_img, voc_dataset.mean_bgr, std=(2.2, 3.5, 4.5))

    decentralized_img = \
        (centralized_img * np.array([2.2, 3.5, 4.5], dtype=np.float32)
         + np.array(voc_dataset.mean_bgr, dtype=np.float32))

    assert np.all(decentralized_img.astype(np.uint8) == pytest.approx(img, abs=1.0))


def test_pad_img_label(voc_dataset):
    """Test pad img and label. The padded size, location and pad value should be correct under following circumstances:
        1. pad size is larger or smaller than img size at H or W
        2. img pad value is number(int, float) or scalar
        3. img pad to size is int or Iterable(H, W)
        4. with or without pad aligner
        5. with or without label
        6. Different pad locations
    Args:
        voc_dataset: VOC Aug val dataset
    """
    print(Fore.LIGHTYELLOW_EX + "===========Test Pad Img Label==========" + Style.RESET_ALL)

    _, img, label = voc_dataset[0]

    def _check_padded_img_label(img, padded_img, padded_h, padded_w, padded_val: np.ndarray,
                                label=None, padded_label=None):
        # * img, label share the same size
        if label is not None:
            assert img.shape[:2] == label.shape
            assert padded_img.shape[:2] == padded_label.shape

        # * padded size is correct
        assert padded_img.shape[0] == padded_h
        assert padded_img.shape[1] == padded_w

        # * left-top part is the same
        assert np.all(padded_img[:img.shape[0], :img.shape[1]] == img)
        if label is not None:
            assert np.all(padded_label[:label.shape[0], :label.shape[1]] == label)

        # * padded value is correct
        img_pad_right = np.zeros(padded_img.shape[:2], dtype=np.bool_)
        img_pad_right[:img.shape[0], :img.shape[1]] = True
        img_pad_right |= np.all(np.isclose(padded_img, padded_val), axis=2)
        assert np.all(img_pad_right)

        if label is not None:
            label_pad_right = np.zeros(padded_label.shape[:2], dtype=np.bool_)
            label_pad_right[:label.shape[0], :label.shape[1]] = True
            label_pad_right |= padded_label == 255
            assert np.all(label_pad_right)

    # * Test diff pad size
    for delta_h in [-10, 10]:
        for delta_w in [-10, 10]:
            pad_img_to = (img.shape[0] + delta_h, img.shape[1] + delta_w)
            img_pad_val = np.array([1, 2, 3])
            padded_img, padded_label = pad_img_label(img, label, pad_img_to, img_pad_val=img_pad_val)
            padded_h = pad_img_to[0] if delta_h > 0 else img.shape[0]
            padded_w = pad_img_to[1] if delta_w > 0 else img.shape[1]
            _check_padded_img_label(img, padded_img, padded_h, padded_w, img_pad_val, label, padded_label)

    # * Test pad int to int size with aligner
    padded_img, padded_label = pad_img_label(img, label, pad_img_to=511, img_pad_val=0,
                                             pad_aligner=find_nearest_odd_size)
    _check_padded_img_label(img, padded_img, 513, 513, np.array([0, 0, 0]), label, padded_label)

    padded_img, padded_label = pad_img_label(img, label, pad_img_to=511, img_pad_val=0,
                                             pad_aligner=[find_nearest_odd_size, find_nearest_even_size])
    _check_padded_img_label(img, padded_img, 513, 512, np.array([0, 0, 0]), label, padded_label)

    # * Test pad float without label
    float_img = int_img2float32_img(img)
    padded_img = pad_img_label(float_img, pad_img_to=513, img_pad_val=np.array([2.3, 3.3, 6.6]))
    assert padded_img.dtype is np.dtype('float32')
    _check_padded_img_label(float_img, padded_img, 513, 513, np.array([2.3, 3.3, 6.6]))

    # * Test Different pad location
    padded_img, padded_label = pad_img_label(img, label, pad_img_to=513, img_pad_val=128, pad_location='left-top')
    assert np.all(padded_img[-img.shape[0]:, -img.shape[1]:] == img)
    assert np.all(padded_label[-img.shape[0]:, -img.shape[1]:] == label)

    padded_img, padded_label = pad_img_label(img, label, pad_img_to=513, img_pad_val=128, pad_location='left-bottom')
    assert np.all(padded_img[:img.shape[0], -img.shape[1]:] == img)
    assert np.all(padded_label[:label.shape[0], -label.shape[1]:] == label)

    padded_img, padded_label = pad_img_label(img, label, pad_img_to=513, img_pad_val=128, pad_location='right-top')
    assert np.all(padded_img[-img.shape[0]:, :img.shape[1]] == img)
    assert np.all(padded_label[-label.shape[0]:, :label.shape[1]] == label)

    padded_img, padded_label = pad_img_label(img, label, pad_img_to=513, img_pad_val=128, pad_location='center')
    pad_h = 513 - img.shape[0]
    pad_w = 513 - img.shape[1]
    assert np.all(padded_img[pad_h // 2:-((pad_h + 1) // 2), pad_w // 2:-((pad_w + 1) // 2)] == img)
    assert np.all(padded_label[pad_h // 2:-((pad_h + 1) // 2), pad_w // 2:-((pad_w + 1) // 2)] == label)


def test_rand_crop(voc_dataset):
    """Test RandCrop
        1. Create RandCrop Auger
        2. Test whether the cropped img and label corresponding to the rand seed
        3. Test error raised when crop size > img size
        4. Test the rand seed distribution is random and balanced
        5. Visualization

    Args:
        voc_dataset: VOCAug val dataset
    """
    print(Fore.LIGHTYELLOW_EX + "===========Test Rand Crop==========" + Style.RESET_ALL)

    kPadSize = 513
    kCropSize = (321, 257)
    kErrorCropSize = (323, 258)

    kLoopTimes = 10000

    class RandCropAuger(BaseAuger):

        def build_graph(self):
            super(RandCropAuger, self).build_graph()
            self.graph.add_node(pad_img_label, inputs=['img', 'label'], outputs=['padded_img', 'padded_label'],
                                kwargs={'pad_img_to': kPadSize, 'img_pad_val': 128})

            self.graph.add_node(RandCrop, inputs=['padded_img', 'padded_label'],
                                outputs=['cropped_img', 'cropped_label'], init={'crop_size': kCropSize})

    auger = RandCropAuger(voc_dataset, verbosity=0, slim=True)
    rand_node: Node = auger.rand_nodes[0]

    def _check_cropped_img_label(img, cropped_img, offset_h, offset_w, label=None, cropped_label=None):
        # * img, label share the same size
        if label is not None:
            assert img.shape[:2] == label.shape
            assert cropped_img.shape[:2] == cropped_label.shape

        # * padded size is correct
        assert cropped_img.shape[0] == kCropSize[0]
        assert cropped_img.shape[1] == kCropSize[1]

        # * cropped part is the same
        assert np.all(img[offset_h: offset_h + kCropSize[0], offset_w: offset_w + kCropSize[1]] == cropped_img)
        if label is not None:
            assert np.all(label[offset_h: offset_h + kCropSize[0], offset_w: offset_w + kCropSize[1]] == cropped_label)

    _, img, label = voc_dataset[0]
    padded_img, padded_label = pad_img_label(img, label, kPadSize, img_pad_val=128)

    offset_h_sum, offset_w_sum = np.zeros(padded_img.shape[0] - kCropSize[0] + 1, dtype=np.int64), \
                                 np.zeros(padded_img.shape[1] - kCropSize[1] + 1, dtype=np.int64)
    for i in range(kLoopTimes):
        cropped_img, cropped_label = auger[0]
        offset_h, offset_w = auger.rand_seeds[rand_node.id]

        offset_h_sum[offset_h] += 1
        offset_w_sum[offset_w] += 1

        _check_cropped_img_label(padded_img, cropped_img, offset_h, offset_w, padded_label, cropped_label)

        if i % 500 == 0:
            print(Fore.LIGHTYELLOW_EX + f"{i}: offset_h = {offset_h}; offset_w = {offset_w}" + Style.RESET_ALL)
            img_pair = [padded_img, cropped_img]
            label_pair = map(label_map2color_map, [padded_label, cropped_label])
            plot_img_label_pair(img_pair, label_pair)

    assert offset_h_sum == pytest.approx(kLoopTimes / (kPadSize - kCropSize[0] + 1), rel=0.5)
    assert offset_w_sum == pytest.approx(kLoopTimes / (kPadSize - kCropSize[1] + 1), rel=0.5)

    auger.graph.add_node(RandCrop, inputs=['cropped_img'], outputs=['wrong_cropped_img'],
                         init={'crop_size': kErrorCropSize})
    with pytest.raises(ValueError, match=f"img_h {kCropSize[0]} must >= crop_h {kErrorCropSize[0]}; " \
                     f"img_w {kCropSize[1]} must >= crop_w {kErrorCropSize[1]}"):
        _ = auger[0]


def test_five_crop(voc_dataset):
    """Test Five Crop with following steps:
        1. Create a FiveCrop auger on voc test dataset
        2. Test whether all img is cropped accroding to it's output_index
        3. Test whether output_index is met with dataset index and auger index
        4. Test whether the total number is right
        5. Visualization

    Args:
        voc_dataset: VOCAug val dataset
    """
    print(Fore.LIGHTYELLOW_EX + "===========Test Rand Crop==========" + Style.RESET_ALL)

    kCropH, kCropW = 513, 321
    kCropSize = (kCropH, kCropW)

    class FiveCropAuger(BaseAuger):

        def build_graph(self):
            super(FiveCropAuger, self).build_graph()

            self.graph.add_node(pad_img_label, inputs=['img', 'label'], outputs=['padded_img', 'padded_label'],
                                kwargs={'pad_img_to': kCropSize, 'img_pad_val': (173, 68, 142),
                                        'pad_location': 'center'})

            self.graph.add_node(FiveCrop, inputs=['padded_img', 'padded_label'],
                                outputs=['cropped_img', 'cropped_label'], init={'crop_size': kCropSize})

    auger = FiveCropAuger(voc_dataset, verbosity=0, slim=True)
    multi_node: Node = auger.multi_nodes[0]

    index_num = np.zeros(5, dtype=np.int64)
    for auger_index, outputs in enumerate(auger):
        dataset_index = auger_index // 5

        cropped_img, cropped_label = outputs
        padded_img, padded_label = auger.graph.data['padded_img'], auger.graph.data['padded_label']
        _, img ,label = voc_dataset[dataset_index]

        output_index = multi_node._fct.output_index
        assert output_index == auger_index % 5

        if output_index == 0:
            assert np.all(cropped_img == padded_img[:kCropH, :kCropW])
            assert np.all(cropped_label == padded_label[:kCropH, :kCropW])
        elif output_index == 1:
            assert np.all(cropped_img == padded_img[:kCropH, -kCropW:])
            assert np.all(cropped_label == padded_label[:kCropH, -kCropW:])
        elif output_index == 2:
            assert np.all(cropped_img == padded_img[-kCropH:, :kCropW])
            assert np.all(cropped_label == padded_label[-kCropH:, :kCropW])
        elif output_index == 3:
            assert np.all(cropped_img == padded_img[-kCropH:, -kCropW:])
            assert np.all(cropped_label == padded_label[-kCropH:, -kCropW:])
        elif output_index == 4:
            head_H = (padded_img.shape[0] - kCropH) // 2
            tail_H = -((padded_img.shape[0] - kCropH + 1) // 2) if head_H != 0 else kCropH
            head_W = (padded_img.shape[1] - kCropW) // 2
            tail_W = -((padded_img.shape[1] - kCropW + 1) // 2) if head_W != 0 else kCropW

            assert np.all(cropped_img == padded_img[head_H:tail_H, head_W:tail_W])
            assert np.all(cropped_label == padded_label[head_H:tail_H, head_W:tail_W])
        else:
            raise ValueError(f"output_index{output_index} not in [0, 1, 2, 3, 4]")

        if dataset_index % 100 == 0:
            print(Fore.BLUE + f"{dataset_index // 100}: output_index = {output_index}\n" + Style.RESET_ALL)
            img_row = [img, padded_img, cropped_img]
            label_row = map(label_map2color_map, [label, padded_label, cropped_label])

            plot_img_label_pair(img_row, label_row)

        index_num[output_index] += 1

    assert np.all(index_num == len(voc_dataset))
    print(Fore.LIGHTYELLOW_EX + f"index_num: {index_num}" + Style.RESET_ALL)


def test_rand_scale(voc_dataset):
    """Test rand scale with following steps:
        1. Create a rand_scale auger on voc val dataset
        2. Test is all img is scale accroding to it's rand_seed
            a) The scaled size is correct
            b) Img and label can be roughly recovered by rescale
        3. Test is rand distribution is random and balanced
        4. Visualization

    Args:
        voc_dataset: VOCAug val dataset
    """
    print(Fore.LIGHTYELLOW_EX + "===========Test Rand Scale==========" + Style.RESET_ALL)

    kScaleFactors = [0.5, 0.75, 1, 1.25, 1.5, 1.75, 2]

    class RandScaleAuger(BaseAuger):

        def build_graph(self):
            super(RandScaleAuger, self).build_graph()

            self.graph.add_node(RandScale, inputs=['img', 'label'], outputs=['scaled_img', 'scaled_label'],
                                init={'scale_factors': kScaleFactors,
                                      'aligner': [lambda x: x + 1, lambda x: x - 1]})

    auger = RandScaleAuger(voc_dataset, verbosity=0, slim=True)
    rand_node: Node = auger.rand_nodes[0]

    def vis():
        img_pair = [img, scaled_img, recovered_img]
        label_pair = map(label_map2color_map, [label, scaled_label, recovered_label])
        plot_img_label_pair(img_pair, label_pair)

    scale_num = np.zeros(len(kScaleFactors), dtype=np.int64)
    img_recover_ratios, label_recover_ratios = [], []
    for index, outputs in enumerate(auger):
        scaled_img, scaled_label = outputs
        assert scaled_img.shape[:2] == scaled_label.shape

        _, img, label = voc_dataset[index]
        img: np.ndarray; label: np.ndarray

        rand_seed = auger.rand_seeds[rand_node.id]
        scale_num[kScaleFactors.index(rand_seed)] += 1

        assert int(img.shape[0] * rand_seed) + 1 == scaled_img.shape[0]
        assert int(img.shape[1] * rand_seed) - 1 == scaled_img.shape[1]

        recovered_img = cv2.resize(scaled_img, img.shape[:2][::-1], interpolation=cv2.INTER_LINEAR)
        img_recover_mask = np.isclose(recovered_img, img, rtol=0.1, atol=5.0)
        img_recover_ratios.append(np.sum(img_recover_mask) / img.size)

        recovered_label = cv2.resize(scaled_label, label.shape[:2][::-1], interpolation=cv2.INTER_NEAREST)
        label_recover_mask = recovered_label == label
        label_recover_ratios.append(np.sum(label_recover_mask) / label.size)

        if img_recover_ratios[-1] < 0.7:
            print(Fore.LIGHTRED_EX +
                  f"img recovery fail. {index}: rand_seed = {rand_seed}\n"
                  + Style.RESET_ALL)
            vis()

        if label_recover_ratios[-1] < 0.7:
            print(Fore.LIGHTRED_EX +
                  f"label recovery fail. {index}: rand_seed = {rand_seed}\n"
                  + Style.RESET_ALL)
            vis()

        if index % 200 == 0:
            print(Fore.LIGHTRED_EX +
                  f"{index // 100}: rand_seed = {rand_seed}\n"
                  + Style.RESET_ALL)
            vis()

    assert np.sum(scale_num) == len(voc_dataset)
    assert scale_num / len(voc_dataset) == pytest.approx(1 / len(kScaleFactors), abs=0.1)

    macro_img_recover_ratio = np.mean(img_recover_ratios)
    macro_label_recover_ratio = np.mean(label_recover_ratios)

    assert macro_img_recover_ratio > 0.7
    assert macro_label_recover_ratio > 0.7

    print(Fore.LIGHTYELLOW_EX +
          f"scale_num = {scale_num}; "
          f"macro_img_recover_ratio = {macro_img_recover_ratio}; "
          f"macro_label_recover_ratio = {macro_label_recover_ratio}"
          + Style.RESET_ALL)


def test_multi_scale(voc_dataset):
    """Test multi scale with following steps:
        1. Create a multi scale auger on voc val dataset
        2. Test whether all img is scaled accroding to it's output_index
            a) The scaled size is correct
            b) Img and label can be roughly recovered by rescale
        3. Test whether output_index is met with dataset index and auger index
        4. Test whether the total number is right
        5. Visualization

    Args:
        voc_dataset: VOCAug val dataset
    """
    print(Fore.LIGHTYELLOW_EX + "===========Test Multi Scale==========" + Style.RESET_ALL)

    kScaleFactors = [0.5, 0.75, 1, 1.25, 1.5, 1.75, 2]

    class MultiScaleAuger(BaseAuger):

        def build_graph(self):
            super(MultiScaleAuger, self).build_graph()

            self.graph.add_node(MultiScale, inputs=['img', 'label'], outputs=['scaled_img', 'scaled_label'],
                                init={'scale_factors': kScaleFactors,
                                      'aligner': [lambda x: x + 1, lambda x: x - 1], 'align_corner': True})

    auger = MultiScaleAuger(voc_dataset, verbosity=0, slim=True)
    multi_node: Node = auger.multi_nodes[0]

    scale_factor_count = [0] * len(kScaleFactors)
    for auger_index, outputs in enumerate(auger):
        scaled_img, scaled_label = outputs
        assert scaled_label.shape == scaled_img.shape[:2]

        dataset_index = auger_index // len(kScaleFactors)
        _, img, label = voc_dataset[dataset_index]

        output_index = multi_node._fct.output_index
        assert output_index == auger_index - len(kScaleFactors) * dataset_index

        scale_factor = multi_node._fct.scale_factors[output_index]
        assert int(img.shape[0] * scale_factor) + 1 == scaled_img.shape[0]
        assert int(img.shape[1] * scale_factor) - 1 == scaled_img.shape[1]

        if dataset_index % 100 == 0:
            print(Fore.BLUE + f"{dataset_index // 100}: scaled_factor = {scale_factor}\n" + Style.RESET_ALL)
            img_pair = [img, scaled_img]
            label_pair = map(label_map2color_map, [label, scaled_label])

            plot_img_label_pair(img_pair, label_pair)

        scale_factor_count[kScaleFactors.index(scale_factor)] += 1

    assert np.sum(scale_factor_count) == len(kScaleFactors) * len(voc_dataset)
    assert scale_factor_count == [len(voc_dataset)] * len(kScaleFactors)
    print(Fore.LIGHTYELLOW_EX + f"scale_factor_count = {scale_factor_count}" + Style.RESET_ALL)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: cnn_align.py
@time: 2020/1/15 6:27
@desc:
"""
__all__ = ["find_nearest_even_size", "find_nearest_odd_size", "divisible_by_n",
           "keep_size_padding", "odd_input_pad_size",
           "even_input_pad_size", "get_q"]


def _check_input(size, min_n: int):
    size = int(size)
    min_n = int(min_n)
    if not isinstance(size, int) or not isinstance(min_n, int):
        raise ValueError(f"size={size}, min_n={min_n} should be able to be int()")
    if size <= 0:
        raise ValueError(f"size=f{size} should > 0")
    if min_n < 1:
        raise ValueError(f"min_n=f{min_n} should >= 1")
    return size, min_n


def find_nearest_odd_size(size, min_n: int=4, is_both_way: bool=False):
    """Find the nearest number to the size which can be writen as k * 2 ^ min_n + 1

    Args:
        size (int): The size of origin img/label
        is_both_way (bool): If True, return odd size nearest number in both way rather only the larger number

    Returns: nearest odd size larger than size if not is_both_way, else nearest odd size smaller or larger than size.
    """
    size, min_n = _check_input(size, min_n)

    base = 2 ** min_n
    k = size // base
    residual = size % base

    if residual == 1 and k >=1:
        return size
    elif not is_both_way:
        return (k * base + 1) if residual == 0 else ((k + 1) * base + 1)
    else:
        left_gap = (base -1) if (residual == 0) else (residual - 1)
        right_gap = 1 if (residual == 0) else (base + 1 - residual)
        return (size - left_gap) if (left_gap < right_gap and (size - left_gap) >= (base + 1)) else (size + right_gap)


def find_nearest_even_size(size, min_n: int=4, is_both_way: bool=False):
    """Find the nearest number to the size which can be writen as k * 2 ^ min_n

    Args:
        size (int): The size of origin img/label
        is_both_way (bool): If True, return even size nearest number in both way rather only the larger number

    Returns: nearest odd size larger than size if not is_both_way, else nearest even size smaller or larger than size.
    """
    size, min_n = _check_input(size, min_n)

    base = 2 ** min_n
    k = size // base
    residual = size % base

    if residual == 0 and k >= 1:
        return size
    elif not is_both_way:
        return (k + 1) * base
    else:
        left_gap = residual
        right_gap = base - residual
        return (size - left_gap) if (left_gap < right_gap and k >= 1) else (size + right_gap)


def divisible_by_n(num: int, n: int, direction='larger', bias: int=0) -> int:
    """Return the largest integer that is divisible by n and less than or equal to num.

    Args:
        num: number
        n: divisor
        direction: 'larger' or 'smaller' or 'nearest', default 'larger'.
        bias: bias to the found number, default 0.

    Returns:
        The bias added larger/smaller/nearest integer that is divisible by n and less than or equal to num.
    """
    match direction:
        case 'larger':
            found = num + (n - num % n) % n
        case 'smaller':
            found = num - num % n
        case 'nearest':
            larger = divisible_by_n(num, n, 'larger')
            smaller = divisible_by_n(num, n, 'smaller')
            found = larger if larger - num <= num - smaller else smaller
        case _:
            raise ValueError(f'Unknown direction: {direction}')
    return found + bias


def get_q(kernel_size: int, dilation: int):
    """Get apparent kernel size

    Args:
        kernel_size: kernel size
        dilation: dilate ratio of kernel
    """
    return (kernel_size - 1) * dilation + 1 if kernel_size % 2 == 1 else kernel_size * dilation


def keep_size_padding(kernel_size: int, dilation: int=1):
    """Return padding size which can remain output size unchanged when stride = 1

    Apparent size of kernel must be an odd lager than 0

    Args:
        kernel_size: Kernel size
        dilation: Dilation ratio for kernel. (Default: 1)

    Returns:
        padding size which can remain output size unchanged when stride = 1 or 2
    """
    q = get_q(kernel_size, dilation)
    if q > 0 and q % 2 == 1:
        return (q - 1) // 2
    else:
        raise ValueError(f"Apparent size of kernel must be an odd lager than 0")


def odd_input_pad_size(kernel_size: int, dilation: int=1):
    """Return proper padding for odd input

    Apparent size of kernel must be an odd lager than 0

    Args:
        kernel_size: Kernel size
        dilation: Dilation ratio for kernel. (Default: 1)

    Returns:
        When stride = 1, return pad size which can remain output size unchanged. When stride = 2, return pad
            size which can make output size = Input Size / 2 + 0.5
    """
    return keep_size_padding(kernel_size, dilation)


def even_input_pad_size(kernel_size: int, dilation: int=1):
    """Return proper padding for even input

    Apparent size of kernel must be an even lager than 0

    Args:
        kernel_size: Kernel size
        dilation: Dilation ratio for kernel

    Returns:
        For stride = 2, return pad size which can make output size = Input Size / 2
    """
    q = get_q(kernel_size, dilation)
    if q > 0 and q % 2 == 0:
        return (q - 2) // 2
    else:
        raise ValueError(f"Apparent size of kernel must be an even lager than 0")


if __name__ == "__main__":
    print(find_nearest_odd_size(1), find_nearest_even_size(2))
    print(find_nearest_odd_size(2), find_nearest_even_size(1))
    print(find_nearest_odd_size(512), find_nearest_odd_size(512, is_both_way=True), find_nearest_even_size(512))
    print(find_nearest_odd_size(513 * 0.75), find_nearest_odd_size(518 * 0.75, is_both_way=True),
          find_nearest_even_size(512 * 0.75))
    print(find_nearest_odd_size(513 * 0.5), find_nearest_odd_size(518 * 0.5, is_both_way=True),
          find_nearest_even_size(512 * 0.5))
    print(find_nearest_odd_size(513 * 1.25), find_nearest_odd_size(518 * 1.25, is_both_way=True),
          find_nearest_even_size(512 * 1.25))
    print(find_nearest_odd_size(513 * 1.5), find_nearest_odd_size(518 * 1.5, is_both_way=True),
          find_nearest_even_size(512 * 1.5))
    print(find_nearest_odd_size(513 * 1.75), find_nearest_odd_size(518 * 1.75, is_both_way=True),
          find_nearest_even_size(512 * 1.75))





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/11/26 21:43
@File    : voc2012_aug2.py
@Software: PyCharm
@Desc    : 第二版的VOC增强器，抛弃了第一版中的静态图Auger。
"""
from functools import partial
from typing import Any, Callable, Iterable

import alchemy_cat.data.plugins.augers as au
import numpy as np
from PIL import Image
from alchemy_cat.acplot import BGR2RGB
from alchemy_cat.alg import size2HW
from alchemy_cat.data import Dataset
from alchemy_cat.py_tools import PackCompose, Compose
from alchemy_cat.dl_config import ADict
from math import ceil
from torchvision.transforms import ToTensor, Normalize, ToPILImage

from .voc2012_auger import lb2cls_lb

kPILMode = Image.BICUBIC  # 适配CLIP。

__all__ = ['VOCAuger']


def _scale_short_size_to(target_size: int, img: np.ndarray, lb: np.ndarray | None=None,
                         aligner: Callable[[int], int] | Iterable[Callable[[int], int]] = lambda x: x,
                         align_corner: bool=False, PIL_mode: int=kPILMode):
    return au.scale_img_label(target_size / min(*img.shape[:2]),
                              img, lb, aligner=aligner,
                              align_corner=align_corner, PIL_mode=PIL_mode)


class VOCAuger(Dataset):

    def __init__(self, dataset: Dataset,
                 is_color_jitter: bool=True,
                 scale_crop_method: Any=None,
                 is_rand_mirror: bool=True,
                 mean: tuple[float, float, float]=(0.48145466, 0.4578275, 0.40821073),
                 std: tuple[float, float, float]=(0.26862954, 0.26130258, 0.27577711),
                 lb_scale_factor: float | None=None, ol_cls_lb: bool=True):
        self.dataset = dataset

        self.is_color_jitter = is_color_jitter
        self.scale_crop_method = scale_crop_method
        self.is_rand_mirror = is_rand_mirror
        self.mean = mean
        self.std = std
        self.lb_scale_factor = lb_scale_factor
        self.ol_cls_lb = ol_cls_lb

        self.rand_color_jitter = au.RandColorJitter()  # VOC参数。

        self.scale_crop: Callable[[np.ndarray, np.ndarray], tuple[np.ndarray, np.ndarray]]
        match scale_crop_method:
            case {'method': 'rand_range',
                  'low_size': low_size, 'high_size': high_size, 'short_thresh': short_thresh,
                  'crop_size': crop_size}:
                assert short_thresh >= crop_size
                self.scale_crop = PackCompose([
                    au.RandRangeScale(low_size, high_size, short_thresh, align_corner=False, PIL_mode=kPILMode),
                    au.RandCrop(crop_size)  # if crop_size is not None else au.pack_identical  # 可以不做crop。
                ])
            case {'method': 'no_scale'} | None:
                self.scale_crop = au.pack_identical
            case (int() | (int(), int())) as size:
                h, w = size2HW(size)
                self.scale_crop = partial(au.scale_img_label, (h, w), align_corner=False, PIL_mode=kPILMode)
            case {'method': 'rand_resize_crop', 'size': size, 'scale': scale, 'ratio': ratio}:
                # 好处：1）面积比例较高时，基本能不丢物体。2）能控制ratio范围。劣势：scale、ratio变化范围有限。
                self.scale_crop = au.RandomResizeCrop(size, scale, ratio, interpolation=kPILMode)
            case {'method': 'scale_align', 'aligner': aligner, 'scale_factors': scale_factors}:  # BS=1下可用。
                # 好处：1）尺寸变化范围任意大，2）不丢失物体；劣势：无法保证ratio不变。
                self.scale_crop = au.MultiScale(scale_factors, aligner, align_corner=False, PIL_mode=kPILMode)
            case {'method': 'scale_long_pad',
                  'low_size': low_size, 'high_size': high_size, 'short_thresh': short_thresh,
                  'aligner': aligner}:
                # 好处：1）尺寸变化范围任意大，2）不丢失物体，3）不改变ratio；劣势：1）需要pad，2）不能隐式地增强相对位置。
                self.scale_crop = PackCompose([
                    au.RandRangeScale(low_size, high_size, short_thresh,
                                      aligner=aligner, align_corner=False, PIL_mode=kPILMode),
                    partial(au.pad_img_label, pad_img_to=high_size,
                            img_pad_val=(np.asarray(mean[::-1]) * 255).astype(np.uint8),
                            ignore_label=255,
                            pad_location='right-bottom', with_mask=True)
                ])
            case {'method': 'fix_short', 'crop_size': crop_size}:
                self.scale_crop = PackCompose([
                    partial(_scale_short_size_to, crop_size),
                    au.RandCrop(crop_size)
                ])
            case {'method': 'fix_short_no_crop', 'target_size': target_size, **others}:
                if others.get('aligner') is not None:
                    self.scale_crop = partial(_scale_short_size_to, target_size, aligner=others['aligner'])
                else:
                    self.scale_crop = partial(_scale_short_size_to, target_size)
            case _:
                raise ValueError(f"不支持的{scale_crop_method=}。")

        self.rand_mirror = au.RandMirror()

        self.to_tensor = ToTensor()
        self.normalize = Normalize(mean, std)

    # @classmethod
    # def train(cls, dataset: Dataset,
    #           scale_crop_method: Any=None,
    #           mean: tuple[int, int, int]=(0.48145466, 0.4578275, 0.40821073),
    #           std: tuple[int, int, int]=(0.26862954, 0.26130258, 0.27577711),
    #           lb_scale_factor: float | None=None):
    #     return cls(dataset=dataset,
    #                is_rand_mirror=True, is_color_jitter=True,
    #                scale_crop_method=scale_crop_method,
    #                mean=mean, std=std,
    #                lb_scale_factor=lb_scale_factor)

    @property
    def inv2PIL(self) -> Compose:
        mean, std = self.mean, self.std
        return Compose([
            Normalize(
                mean=[-mean[0] / std[0], -mean[1] / std[1], -mean[2] / std[2]],
                std=[1 / std[0], 1 / std[1], 1 / std[2]]
            ),
            ToPILImage('RGB')
        ])

    @classmethod
    def test(cls, dataset: Dataset,
             mean: tuple[int, int, int]=(0.48145466, 0.4578275, 0.40821073),
             std: tuple[int, int, int]=(0.26862954, 0.26130258, 0.27577711),
             ):
        return cls(dataset=dataset,
                   is_rand_mirror=False, is_color_jitter=False,
                   scale_crop_method=None,
                   mean=mean, std=std,
                   lb_scale_factor=None, ol_cls_lb=False)

    def get_item(self, index) -> ADict:
        inp = self.dataset[index]
        img_id, img, lb, cls_lb = inp.img_id, inp.img, inp.lb, inp.cls_lb

        out = ADict()

        # * 色彩抖动。
        if self.is_color_jitter:
            img = self.rand_color_jitter(img)
        # * 缩放。
        if len(scale_crop_out := self.scale_crop(img, lb)) == 2:
            img, lb = scale_crop_out
        else:
            img, lb, padding_mask, pad_pos = scale_crop_out
            out.pad_info.padding_mask, out.pad_info.pad_pos = padding_mask, pad_pos
        # * 随机镜像。
        if self.is_rand_mirror:
            img, lb = self.rand_mirror(img, lb)
        # * 获取下采样后lb。
        if self.lb_scale_factor is not None:
            # 训练时，缩放裁剪后，图片尺寸是16k或16k+1，缩放后应当为k和k+1，ceil满足要求。
            scaled_lb = au.PIL2arr(au.arr2PIL(lb).resize((ceil(lb.shape[1] / self.lb_scale_factor),
                                                          ceil(lb.shape[0] / self.lb_scale_factor)),
                                                         resample=Image.NEAREST))
            out.scaled_lb = scaled_lb.astype(np.int64)
        # * 获取增强后标签中类别。
        if self.ol_cls_lb:
            out.ol_cls_lb = lb2cls_lb(lb).astype(np.int64)  # TODO 令该函数所有数据集通用。

        # * 测试增强。
        img = self.to_tensor(BGR2RGB(img).copy())
        img = self.normalize(img)

        # * 返回结果。
        out.img_id, out.img, out.lb, out.cls_lb = img_id, img, lb.astype(np.int64), cls_lb.astype(np.int64)
        return out

    def __len__(self):
        return len(self.dataset)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/27 14:08
@File    : cfg2tune_runner.py
@Software: PyCharm
@Desc    : 
"""
import os.path as osp
import subprocess
from functools import wraps
from multiprocessing import Pool
from pprint import pprint
from typing import List, Optional, Any, Callable

import pandas as pd
from colorama import Style, Fore
from openpyxl import load_workbook
from tqdm import tqdm

from ..logger import Logger
from ..str_formatters import get_local_time_str
from .cfg2tune import Cfg2Tune
from ..config.py_cfg import Config

__all__ = ["Cfg2TuneRunner"]


class Cfg2TuneRunner(object):
    """Running cfg2tune"""
    def __init__(self, cfg2tune_py: str, config_root: str='./configs', experiment_root="experiment", pool_size: int=0,
                 metric_names: Optional[List[str]]=None,
                 gather_metric_fn: Callable[[Config, str, Any, dict[str, tuple[Any, str]]], dict[str, Any]]=None,
                 work_fn: Callable[[tuple[int, tuple[Config, str, str]]], Any]=None):
        """Running a cfg2tune.

        Args:
            cfg2tune_py: cfg2tune python script.
            config_root: Root dir where configs store.
            experiment_root: Root dir where experiment results store.
            pool_size: Multiprocess' Pool size when running tuning work.
            metric_names: Name of metrics to be gathered.
        """
        self.cfg2tune_py = cfg2tune_py
        self.config_root = config_root
        self.experiment_root = experiment_root
        self.pool_size = pool_size
        self.gather_metric_fn = gather_metric_fn
        self.work_fn = work_fn

        # * 加载Cfg2Tune。
        self.cfg2tune = Cfg2Tune.load_cfg2tune(cfg2tune_py, config_root)

        # * 设置整个调参实验的结果文件夹。
        self.rslt_dir = osp.join(experiment_root, self.cfg2tune.rslt_dir)

        # * 获取指标名。
        if self.cfg2tune.metric_names:
            assert metric_names is None
            self.metric_names = self.cfg2tune.metric_names
        else:
            self.metric_names = metric_names

        # * 所有参数组合+参数组合对应的metric。
        self.param_combs: List[dict[str, tuple[Any, str]]] = self.cfg2tune.param_combs
        self.metric_frame: Optional[pd.DataFrame] = None

        # * 每个配置的pkl及其对应的实验文件夹。
        self.cfg_pkls: List[str] = []
        self.cfgs: list[Config] = []
        self.cfg_rslt_dirs: List[str] = []

        # * 保存每个配置的运行结果。
        self.run_rslts: List[subprocess.CompletedProcess] = []

    def build_metrics(self):
        """得到参数组合与metric frame。"""
        midx = pd.MultiIndex.from_tuples([tuple(val[1] for val in param_comb.values())
                                          for param_comb in self.param_combs],
                                         names=list(self.param_combs[0].keys()))
        self.metric_frame = pd.DataFrame(index=midx, columns=self.metric_names)

    def set_cfgs(self):
        """将Cfg2Tune转存为pkl文件，并得到对应的实验文件夹。"""
        self.cfg_pkls, self.cfgs = self.cfg2tune.dump_cfgs(self.config_root)
        self.cfg_rslt_dirs = [osp.join(self.rslt_dir, osp.basename(osp.dirname(cfg_pkl))) for cfg_pkl in self.cfg_pkls]

    def run_cfgs(self):
        """并行或串行地，根据每个配置，执行work函数。work函数内，应当完成一次实验。"""
        work = self.work_fn if self.work_fn is not None else self.work

        if self.pool_size > 0:
            with Pool(self.pool_size) as p:
                map_it = p.imap(work, enumerate(zip(self.cfgs, self.cfg_pkls, self.cfg_rslt_dirs)), chunksize=1)
                for run_rslt in tqdm(map_it, 'Tuning', len(self.cfg_pkls), unit='configs', dynamic_ncols=True):
                    self.run_rslts.append(run_rslt)
        else:
            for pkl_idx, cfg_cfg_pkl_cfg_rslt_dir in tqdm(enumerate(zip(self.cfgs, self.cfg_pkls, self.cfg_rslt_dirs)),
                                                           'Tuning', len(self.cfg_pkls),
                                                           unit='configs', dynamic_ncols=True):
                try:
                    run_rslt = work((pkl_idx, cfg_cfg_pkl_cfg_rslt_dir))
                except subprocess.CalledProcessError as e:
                    print(f"Cfg2TuneRunner's work failed. The stdout and stderr of work are: ")
                    print(f"{e.stdout}\n\n{e.stderr}")
                    raise e
                self.run_rslts.append(run_rslt)

    @staticmethod
    def parse_work_param(pkl_idx_cfg_cfg_pkl_cfg_rslt_dir: tuple[int, tuple[Config, str, str]])\
            -> tuple[int, Config, str, str]:
        """pkl_idx, (cfg, cfg_pkl, cfg_rslt_dir) = pkl_idx_cfg_cfg_pkl_cfg_rslt_dir"""
        pkl_idx, (cfg, cfg_pkl, cfg_rslt_dir) = pkl_idx_cfg_cfg_pkl_cfg_rslt_dir
        return pkl_idx, cfg, cfg_pkl, cfg_rslt_dir

    @staticmethod
    def work(pkl_idx_cfg_cfg_pkl_cfg_rslt_dir: tuple[int, tuple[Config, str, str]]):
        """pkl_idx, (cfg, cfg_pkl, cfg_rslt_dir) = pkl_idx_cfg_cfg_pkl_cfg_rslt_dir,
        then run the config pkl with subprocess.

        Return subprocess.CompletedProcess. """
        raise NotImplementedError()

    def register_work_fn(self, work_fn: Callable[[int, Config, str, str], Any]) \
            -> Callable[[tuple[int, tuple[Config, str, str]]], Any]:
        """Register work_fn. Use as a decorator."""
        @wraps(work_fn)
        def wrapper(pkl_idx_cfg_cfg_pkl_cfg_rslt_dir: tuple[int, tuple[Config, str, str]]):
            pkl_idx, cfg, cfg_pkl, cfg_rslt_dir = Cfg2TuneRunner.parse_work_param(pkl_idx_cfg_cfg_pkl_cfg_rslt_dir)
            return work_fn(pkl_idx, cfg, cfg_pkl, cfg_rslt_dir)

        self.work_fn = wrapper

        return wrapper

    def gather_metrics(self):
        for cfg, cfg_rslt_dir, run_rslt, param_comb in zip(self.cfgs, self.cfg_rslt_dirs,
                                                           self.run_rslts, self.param_combs):
            metric = self.gather_metric(cfg, cfg_rslt_dir, run_rslt, param_comb)

            if self.metric_frame is None:
                if self.metric_names is None:
                    self.metric_names = list(metric.keys())
                self.build_metrics()
            else:
                assert set(metric.keys()) == set(self.metric_names), \
                    f"metric.keys()={metric.keys()} != self.metric_names={self.metric_names}"

            self.metric_frame.loc[tuple(val[1] for val in param_comb.values())] = metric

    def gather_metric(self, cfg: Config, cfg_rslt_dir: str, run_rslt: Any, param_comb: dict[str, tuple[Any, str]]) \
            -> dict[str, Any]:
        """Given cfg_rslt_dir, run_stdout, param_comb, return {metric_name: metric}"""
        if self.gather_metric_fn is not None:
            return self.gather_metric_fn(cfg, cfg_rslt_dir, run_rslt, param_comb)
        else:
            raise NotImplementedError("gather_metric not implemented")

    def register_gather_metric_fn(self,
                                  gather_metric_fn: Callable[[Config, str, Any, dict[str, tuple[Any, str]]],
                                                             dict[str, Any]]):
        """Register gather_metric_fn. Use as a decorator."""
        self.gather_metric_fn = gather_metric_fn
        return gather_metric_fn

    def save_metrics(self):
        exp_short_id = osp.basename(osp.dirname(self.cfg2tune_py)).split('@')[-1]
        # 美观起见，excel选择merge cells。需要时可以手动解除merge。
        self.metric_frame.to_excel(osp.join(self.rslt_dir, f'{exp_short_id}.xlsx'), sheet_name='MetricFrame',
                                   merge_cells=True)
        # 将刚才保存的xlsx转储为xlsm，用于数据分析。
        load_workbook(filename=osp.join(self.rslt_dir, f'{exp_short_id}.xlsx'),
                      read_only=False, keep_vba=True).save(osp.join(self.rslt_dir, f'{exp_short_id}.xlsm'))
        self.metric_frame.to_pickle(osp.join(self.rslt_dir, f'{exp_short_id}.pkl'))

    def tuning(self):
        stdout_log_file = osp.join(self.rslt_dir, 'stdout',
                                   '-'.join(['stdout', get_local_time_str(for_file_name=True)]) + '.log')
        Logger(stdout_log_file, real_time=True)
        print(f"Log stdout at {stdout_log_file}")

        print(Fore.GREEN + "-----------------Setting Configs-----------------" + Style.RESET_ALL)
        self.set_cfgs()

        print("Saving Config pkls at: ")
        pprint(self.cfg_pkls)

        print("Param Combines: ")
        pprint(self.param_combs, sort_dicts=False)

        print(Fore.GREEN + "-----------------Running Configs-----------------" + Style.RESET_ALL)
        self.run_cfgs()

        if self.gather_metric_fn is not None:
            print(Fore.GREEN + "-----------------Gather Metrics-----------------" + Style.RESET_ALL)
            self.gather_metrics()
            print("Metric Frame: ")
            pprint(self.metric_frame)

            print(Fore.GREEN + "-----------------Save Metrics-----------------" + Style.RESET_ALL)
            self.save_metrics()
            print(f"Saving Metric Frame at {osp.join(self.rslt_dir, 'metric_frame.xlsx')}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: __init__.py
@time: 2019/12/7 23:58
@desc:
"""
from .cnn_align import *
from .utils import *
from .msc_flip_inference import msc_flip_inference
from .clamp_softmax import *
# NOTE data初始化调用alg，alg的dense_crf调用data.plugin, data.plugin调用data.datasets，导致循环import。
# from alchemy_cat.alg.dense_crf import *
from .normalize_tensor import *
from .complement_entropy_confidence import *
from .slide_inference import *
from .window_slides import *
from .masked_softmax import *
from .resize import *
from .sampling import *
# NOTE 同densecrf情况，eval_cam的导入会触发data模块的循环导入。
# from .eval_cams import *



#!/usr/bin/env src
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: e2e.py
@time: 2020/3/23 4:23
@desc:
"""
from typing import Dict, Tuple
from collections import defaultdict
import re

import torch
from torch import nn

from alchemy_cat.contrib.models.wsss.vgg_feature import VGGFeature
from alchemy_cat.contrib.models.wsss.vgg_ASPP import VGGASPP, FCReLUDrop
from alchemy_cat.contrib.models.wsss.pam import PAM

from alchemy_cat.torch_tools import Network
from alchemy_cat.torch_tools import strip_param_name, strip_named_params, update_model_state_dict

__all__ = ['VGGE2ECls', 'VGGE2EPamCollab', 'E2EVGGBackbone', 'E2EClsHead', 'E2EASPPSegHead', 'E2EPAMSegHead']


class E2EVGGBackbone(VGGFeature, Network):
    """VGG Backbone for end to end module for WSSS"""
    def __init__(self):
        super(E2EVGGBackbone, self).__init__(in_ch=3, out_ch=[64, 128, 256, 512, 512],
                                             conv_nums=[2, 2, 3, 3, 3], dilations=[1, 1, 1, 1, 2],
                                             pool_strides=[2, 2, 2, 1, 1], pool_size=[3, 3, 3, 1, 1])


class E2EClsHead(nn.Sequential, Network):
    """Classification head for end to end module for WSSS"""
    def __init__(self, in_ch: int, num_class: int, branch_idx: int):
        super(E2EClsHead, self).__init__()

        self.in_channel = in_ch
        self.out_channel = num_class

        self.add_module(f"e2e_cls_head6_{branch_idx}", FCReLUDrop(in_ch, 1024, kernel_size=3, dilation=1, padding=1,
                                                                  layer_idx=6, branch_idx=branch_idx))
        self.add_module(f"e2e_cls_head7_{branch_idx}", FCReLUDrop(1024, 1024, kernel_size=1, dilation=1, padding=0,
                                                                  layer_idx=7, branch_idx=branch_idx))
        self.add_module(f"fc8_{branch_idx}", nn.Conv2d(1024, num_class, 1, bias=False))

    def initialize(self, pretrain_dict: Dict[str, nn.Parameter], verbosity: int=1, **kwargs):
        vgg_init_from = kwargs.get('vgg_init_from', 'deeplab_v2')

        if vgg_init_from == 'deeplab_v2':
            self.initialize_from_deeplab_v2(pretrain_dict, verbosity)
        elif vgg_init_from == 'deeplab_v1':
            self.initialize_from_deeplab_v1(pretrain_dict, verbosity)
        else:
            raise ValueError(f"kwarg vgg_init_from = {vgg_init_from} must be deeplab_v1 or deeplab_v2.")

    def initialize_from_deeplab_v2(self, pretrain_dict: Dict[str, nn.Parameter], verbosity: int=1):
        strip_pretrain_dict = strip_named_params(pretrain_dict)
        weight_dict = {}

        for model_param_name in self.state_dict().keys():
            strip_model_param_name = strip_param_name(model_param_name)
            if strip_model_param_name in strip_pretrain_dict:
                # Exact Match
                weight_dict[model_param_name] = strip_pretrain_dict[strip_model_param_name]
            else:
                if 'fc8' in model_param_name:
                    # Don't init fc_8
                    continue
                # Fuzzy Match
                fuzzy_strip_model_param_name = re.sub(r'fc\d_\d',
                                                      lambda match: match.group(0)[:-1] + '1', strip_model_param_name)
                weight_dict[model_param_name] = strip_pretrain_dict[fuzzy_strip_model_param_name]

        update_model_state_dict(self, weight_dict, verbosity)

    def initialize_from_deeplab_v1(self, pretrain_dict: Dict[str, nn.Parameter], verbosity: int=1):
        update_dict = {}
        matched_dict, missed_keys = self.find_match_miss(pretrain_dict)
        update_dict.update(matched_dict)

        if len(missed_keys) > 0:
            print("Init cls head from deeplab_v1")
            strip_pretrain_dict = strip_named_params(pretrain_dict)
            for model_param_name in missed_keys:
                strip_model_param_name = strip_param_name(model_param_name)
                if 'fc8' in strip_model_param_name:
                    continue
                else:
                    param_name_in_pretrain = re.sub(r'(fc\d)_\d', lambda match: match.group(1), strip_model_param_name)
                    update_dict[model_param_name] = strip_pretrain_dict[param_name_in_pretrain]

        self.update_model_state_dict(update_dict, verbosity)


class E2EASPPSegHead(VGGASPP, Network):
    """Segmentation head with ASPP for end to end module for WSSS"""
    def __init__(self, in_ch: int, num_class: int):
        super(E2EASPPSegHead, self).__init__(in_ch, num_class,
                                             rates=[6, 12, 18, 24], start_layer_idx=6)


class E2EPAMSegHead(nn.Sequential, Network):
    """Segmentation head with PAM for end to end module for WSSS"""
    def __init__(self, in_ch: int, num_class: int):
        super(E2EPAMSegHead, self).__init__()

        self.in_channel = in_ch
        self.out_channel = num_class

        self.pre_pam = nn.Sequential(
            FCReLUDrop(in_ch, 1024, kernel_size=3, dilation=6, padding=6, layer_idx=6, branch_idx=1),
            FCReLUDrop(1024, 1024, kernel_size=1, dilation=1, padding=0, layer_idx=7, branch_idx=1)
        )

        self.pam = PAM(1024, 256, num_class, layer_idx=8, branch_idx=1)

    def initialize(self, pretrain_dict: Dict[str, nn.Parameter], verbosity: int=2, **kwargs):
        weight_dict = dict()
        matched_dict, missed_keys = self.find_match_miss(pretrain_dict)
        weight_dict.update(matched_dict)

        if len(missed_keys) > 0:
            print("Initial seg head from cls head")
            strip_pretrain_dict = strip_named_params(pretrain_dict)
            for model_param_name in missed_keys:
                if 'fc8' in model_param_name:
                    # Don't init fc_8
                    continue
                if re.search(r'fc\d_\d', model_param_name) is None:
                    continue
                # Fuzzy Match
                strip_model_param_name = strip_param_name(model_param_name)
                fuzzy_strip_model_param_name = re.sub(r'(fc\d_)\d', lambda match: match.group(1) + '0',
                                                      strip_model_param_name)
                weight_dict[model_param_name] = strip_pretrain_dict[fuzzy_strip_model_param_name]

        self.update_model_state_dict(weight_dict, verbosity)


class _E2EBase(Network):
    """Define get_params for all e2e model"""
    def get_param_groups(self) -> dict:
        param_groups = defaultdict(list)
        for name, param in self.named_parameters():
            if 'backbone' in name:
                if 'weight' in name:
                    param_groups['backbone_weight'].append(param)
                elif 'bias' in name:
                    param_groups['backbone_bias'].append(param)
                else:
                    raise RuntimeError(f"UnKnown param {name}: {param}")
            elif 'head' in name:
                if 'weight' in name:
                    param_groups['head_weight'].append(param)
                elif 'bias' in name:
                    param_groups['head_bias'].append(param)
                elif 'gamma' in name:
                    param_groups['head_gamma'].append(param)
                else:
                    raise RuntimeError(f"Unknown param {name}: {param}")
            else:
                raise RuntimeError(f"Unknown param {name}: {param}")

        return dict(param_groups)


class VGGE2ECls(nn.Sequential, _E2EBase, Network):
    """End to end module's classification branch for WSSS"""

    def __init__(self, num_class: int=21):
        super(VGGE2ECls, self).__init__()

        self.in_channel = 3
        self.out_channel = num_class - 1

        self.backbone = VGGFeature(in_ch=3, out_ch=[64, 128, 256, 512, 512],
                                   conv_nums=[2, 2, 3, 3, 3], dilations=[1, 1, 1, 1, 2],
                                   pool_strides=[2, 2, 2, 1, 1], pool_size=[3, 3, 3, 1, 1])
        self.cls_head = E2EClsHead(self.backbone.out_channel, num_class-1, branch_idx=0)


class VGGE2EPamCollab(_E2EBase, Network):
    """Collab End to End model with pad seg head for WSSS"""

    def __init__(self, num_class: int=21):
        super().__init__()

        self.in_channel = 3

        self.vgg_e2e_cls = VGGE2ECls(num_class)

        self.seg_head = E2EPAMSegHead(self.vgg_e2e_cls.backbone.out_channel, num_class)

    def forward(self, img) -> Tuple[torch.Tensor, torch.Tensor]:
        feature_map = self.vgg_e2e_cls.backbone(img)
        cam = self.vgg_e2e_cls.cls_head(feature_map)
        score_map = self.seg_head(feature_map)

        return cam, score_map



#!/usr/bin/env src
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: vgg_ASPP.py
@time: 2019/12/12 16:02
@desc:
"""
from typing import Dict

import re

import torch
import torch.nn as nn

from alchemy_cat.torch_tools import strip_named_params, strip_param_name, update_model_state_dict, named_param_match_pattern
from alchemy_cat.torch_tools import Network

__all__ = ['VGGASPPBranch', 'VGGASPP', 'FCReLUDrop']


class FCReLUDrop(nn.Sequential, Network):

    def __init__(self, in_ch, out_ch, kernel_size, dilation, padding, layer_idx, branch_idx):
        super(FCReLUDrop, self).__init__()

        self.in_channel = in_ch
        self.out_channel = out_ch

        self.add_module(f"fc{layer_idx}_{branch_idx}",
                        nn.Conv2d(in_ch, out_ch, kernel_size, stride=1, padding=padding, dilation=dilation))

        self.add_module(f"relu{layer_idx}_{branch_idx}",
                        nn.ReLU(inplace=True))

        self.add_module(f"drop{layer_idx}_{branch_idx}",
                        nn.Dropout(p=0.5))


class VGGASPPBranch(nn.Sequential, Network):

    def __init__(self, in_ch, num_classes, rate, start_layer_idx, branch_idx, net_id):
        super(VGGASPPBranch, self).__init__()

        self.in_channel = in_ch
        self.out_channel = num_classes

        self.add_module(f"aspp_layer{start_layer_idx}_{branch_idx}",
                        FCReLUDrop(in_ch, out_ch=1024, kernel_size=3, dilation=rate, padding=rate,
                                   layer_idx=start_layer_idx, branch_idx=branch_idx))

        self.add_module(f"aspp_layer{start_layer_idx + 1}_{branch_idx}",
                        FCReLUDrop(in_ch=1024, out_ch=1024, kernel_size=1, dilation=1, padding=0,
                                   layer_idx=start_layer_idx + 1, branch_idx=branch_idx))

        self.add_module(f"fc{start_layer_idx + 2}_{net_id}_{branch_idx}",
                        nn.Conv2d(in_channels=1024, out_channels=num_classes, kernel_size=1))

        # for module in self.children():
        #     if isinstance(module, nn.Conv2d):
        #         nn.init.normal_(module.weight, mean=0.0, std=0.01)
        #         nn.init.constant_(module.bias, 0.0)
        fc_logit = getattr(self, f"fc{start_layer_idx + 2}_{net_id}_{branch_idx}")
        nn.init.normal_(fc_logit.weight, mean=0.0, std=0.01)
        nn.init.constant_(fc_logit.bias, 0.0)

    def initialize(self, pretrain_dict: Dict[str, nn.Parameter], verbosity: int=1, **kwargs):
        strip_pretrain_dict = strip_named_params(pretrain_dict)
        weight_dict = {}

        for model_param_name in self.state_dict().keys():
            strip_model_param_name = strip_param_name(model_param_name)
            if strip_model_param_name in strip_pretrain_dict:
                weight_dict[model_param_name] = strip_pretrain_dict[strip_model_param_name]
            elif 'fc8' in strip_model_param_name:
                continue
            else:
                # Using param from cls branch with same layer_ID
                print(f"Init {strip_model_param_name} from cls head")
                pattern = re.sub(r'fc\d_\d', lambda match: match.group(0)[:-1] + '0', strip_model_param_name)
                weight = list(named_param_match_pattern(pattern,
                                                        strip_pretrain_dict,
                                                        full_match=True,
                                                        multi_match=False).values())[0]
                weight_dict[model_param_name] = weight
        update_model_state_dict(self, weight_dict, verbosity)


class VGGASPP(Network):

    def __init__(self, in_ch, num_classes, rates, start_layer_idx, branch_indices, net_id="pascal"):
        super(VGGASPP, self).__init__()

        self.in_channel = in_ch
        self.out_channel = num_classes

        for rate, branch_idx in zip(rates, branch_indices):
            self.add_module(f"aspp_branch{branch_idx}",
                            VGGASPPBranch(in_ch, num_classes, rate, start_layer_idx, branch_idx, net_id))

    def forward(self, x):
        return sum([branch(x) for branch in self.children()])


if __name__ == "__main__":
    aspp = VGGASPP(512, 21, [6, 12, 18, 24], 6, range(1, 5))

    for name, module in aspp.named_modules():
        if "fc8" in name:
            print("-----------------------------------")
            print(name)
            print(f"Weight std of module is {torch.std(module.weight)}, "
                  f"Weight mean of module is {torch.mean(module.weight)}")
            print(f"Bias sum of the module is {torch.sum(module.bias)}")

    from torch.utils.tensorboard import SummaryWriter
    writer = SummaryWriter(log_dir="temp")

    writer.add_graph(aspp, input_to_model=torch.randn((10, 512, 41, 41), dtype=torch.float32))

    print("Graph has been writen to the temp dir")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: tracker.py
@time: 2020/2/26 13:40
@desc:
"""
import json
import os
import pickle
from pprint import pformat, pprint
from typing import Any, Callable, DefaultDict, Optional
from collections import defaultdict, OrderedDict
import warnings

from alchemy_cat.py_tools import indent
from alchemy_cat.py_tools.numpy_json_encoder import NumpyArrayEncoder

__all__ = ['Statistic', 'Tracker', 'OneOffTracker']


class _StatisticStatue(object):

    def __init__(self):
        self.last_update_count: int = -1
        self.last_statistic: Any = None


def _check_importance(importance):
    if not isinstance(importance, int) or importance < 0 or importance > 9:
            raise ValueError(f"importance {importance} must >= 0 and <= 9")


class Statistic(object):
    """Decorator for statistic of Tracker

    Function decorated should specify how to calculate the statistic from Tracker object. The Statistic obj can
    caching the last statistic and return it if Tracker is not updated since last calculate.

    You can also use Statistic.getter to create statistics with importance, which can be filtered when get
    self.statistics
    """

    def __init__(self, fget: Callable, importance: int=0):
        self.fget = fget
        self.name = fget.__name__
        self.__doc__ = fget.__doc__

        _check_importance(importance)
        self.importance = importance

    @classmethod
    def getter(cls, importance: int=0):
        """Decorate statics with importance"""
        _check_importance(importance)

        def decorator(fget):
            return cls(fget, importance)

        return decorator

    def __get__(self, obj: 'Tracker', obj_type=None):
        if obj is None:
            return self

        statistic_statue: _StatisticStatue = obj._statistic_status[self.name]

        if statistic_statue.last_update_count < obj._update_count:
            # * Calculate statistic
            statistic_statue.last_update_count = obj._update_count
            statistic_statue.last_statistic = self.fget(obj)
            return statistic_statue.last_statistic
        elif statistic_statue.last_update_count > obj._update_count:
            # * Raise Error
            raise RuntimeError(f"Statistic {self.name}'s last_update_count {statistic_statue.last_update_count} can't "
                               f"larger than tracker {obj}'s update_count {obj.update_count}")
        else:
            # * Return cache
            return statistic_statue.last_statistic

    def __repr__(self):
        return f"Statistic <{self.__class__}>: {self.name}"


class Tracker(object):
    """A tracker to some progress.

    Tracker object is used to tracking some progress. It can be updated with progress and calculate statistic
    of the tracked progress.
    """

    def __init__(self, init_dict: Optional[OrderedDict]=None):
        """Init tracker's attributes

        Must be called before first in custom __init__.

        Args:
            init_dict: init_dict for reset the tracker. Make sure to call by super().__init__(quick_init(self, locals())
                if subclass is directly driven from Tracker, other wise call
                Tracker.__init__(self, quick_init(self, locals()) in last driven class's __init__ or just modify
                self.init_dict If None, then can't use self.reset.
        """
        self._update_count: int = 0
        self._statistic_status: DefaultDict[str, _StatisticStatue] = defaultdict(_StatisticStatue)
        self.init_dict = init_dict

    def update(self, *args, **kwargs):
        """Update tracker's statues

        Must be called first in custom update
        """
        if hasattr(self, 'is_one_off') and self.is_one_off and not self.is_in_one_off_block:
            warnings.warn(f"You are Updating the one-off tracker {self} out-off the with block", RuntimeWarning)
        self._update_count += 1

    @property
    def update_count(self):
        """Return update count of current tracker"""
        return self._update_count

    def statistics(self, importance: int=0) -> OrderedDict:
        """Return all statistic of current tracker in the dict {statistic.name: statistic}

        Args:
            importance (int): importance filter. Only statistics with (statistics.importance > importance) will be
                returned
        """
        _check_importance(importance)

        tracker_classes = [base_cls for base_cls in self.__class__.mro() if issubclass(base_cls, Tracker)]

        statistics = []
        statistic_names = []
        for cls in tracker_classes:
            for statistic in cls.__dict__.values():
                # Make sure that on class's statistic can override base class's statistic
                if isinstance(statistic, Statistic) and statistic.name not in statistic_names:
                    statistics.append(statistic)
                    statistic_names.append(statistic.name)

        statistic_dict = {statistic: statistic.__get__(self, self.__class__)
                for statistic in statistics if statistic.importance >= importance}

        sorted_statistics = sorted(statistic_dict.items(), key=lambda item: str(item[0].importance) + item[0].name)

        ordered_statistic = OrderedDict()
        for k, v in sorted_statistics:
            ordered_statistic[k.name] = v

        return ordered_statistic

    def save_statistics(self, save_dir: str, importance: int=0):
        """Save statistics needing saving to save_dir.

        If self.statistic can be saved by json, then function will save statistics to statistics.json, else to
        statistics.txt. self.statistics will be pickle to statistic.pkl anyway.

        Args:
            save_dir: Dictionary to save the statistic.json/txt and statistic.pkl
            importance (int): importance filter. Only statistics with statistics.importance > importance will be saved
        """
        _check_importance(importance)
        os.makedirs(save_dir, exist_ok=True)

        # * Save statistics
        try:
            with open(os.path.join(save_dir, 'statistics.json'), 'w') as json_f:
                json.dump(self.statistics(importance), json_f, indent=4, cls=NumpyArrayEncoder)
        finally:
            with open(os.path.join(save_dir, 'statistics.txt'), 'w') as txt_f:
                txt_f.write(pformat(self.statistics(importance), indent=4))

            with open(os.path.join(save_dir, 'statistics.pkl'), 'wb') as pkl_f:
                pickle.dump(self.statistics(importance), pkl_f)

    def print_statistics(self, importance: int=0):
        """Print statistics filtered by importance

        Args:
            importance: importance filter. Only statistics with statistics.importance > importance will be saved

        Returns:
            Formatted statistics filtered by importance
        """
        _check_importance(importance)

        pprint(self.statistics(importance))
        return pformat(self.statistics(importance), indent=4)

    def reset(self):
        if self.init_dict is None:
            raise RuntimeError(f"Tracker {self} has no init_dict. Set init_dict with quick_init in __init__")

        return self.__init__(**self.init_dict)

    def __repr__(self):
        return f"Tracker <{self.__class__}>: update_count = {self.update_count}"

    def __str__(self):
        return pformat(self.statistics(), indent=4)


class OneOffTracker(object):
    """An context manager which to make sure that tracker can only be updated within the context"""

    def __init__(self, tracker_factory: Callable[[], Tracker]):
        """An context manager which to make sure that tracker can only be updated within the context

        Args:
            tracker_factory: Return the one-off tracker
        """
        self._tracker = tracker_factory()
        self._tracker.is_one_off = True
        self._tracker.is_in_one_off_block = False

    def __enter__(self):
        self._tracker.is_in_one_off_block = True
        return self._tracker

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._tracker.is_in_one_off_block = False

    def __repr__(self):
        return f"OneOffTracker <{self.__class__}>: \n" \
            + indent(f"tracker: {self._tracker}\n") \
            + indent(f"tracker.is_in_one_off_block: {self._tracker.is_in_one_off_block}")


# TODO: Trigger decorator which can be trigger after function called several times.



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/5/23 22:54
@File    : load_module.py
@Software: PyCharm
@Desc    : 
"""
from types import ModuleType

import os.path as osp
from importlib import import_module
from importlib.util import spec_from_file_location, module_from_spec
# import traceback
import warnings

__all__ = ['load_module_from_py']


def load_module_from_py(py: str, from_file: bool=False) -> ModuleType:
    """给定python文件路径，载入python模块。

    函数首先会尝试将python文件路径转换为'A.B.C'的导入路径，随后用标准的import_module函数导入文件为模块。
    若失败（譬如原路径中含有“.”字符），就尝试直接利用文件查找器+载入器，将python文件载入为模块，并名以“foo”，随后执行文件，绑定
    模块命名空间中对象。

    应当注意，尽量避免进入第二种导入方式。以导入链A import B、C，B、C import D为例：

    1. 文件直接导入不会将模块加入sys.module中，模块导完就释放。若某模块需要导入两次（如上例中D），则D需要独立导入两次。不仅浪费，
    且B、C导入的D中同名对象，并非真正的同一对象（如某些容器、模型），这可能会带来我们不希望的行为。

    2. 文件直接导入所得模块，其中函数被pickle转储时，函数地址（func.__module__）为无意义的foo。故载入时，无法根据地址找到被转储
    的原函数。

    Args:
        py: python文件路径。
        from_file: 是否强制从文件导入。

    Returns:
        载入的python模块。
    """
    module = None

    if not from_file:
        try:
            import_path = '.'.join(osp.normpath(osp.splitext(py)[0]).lstrip(osp.sep).split(osp.sep))
            if '.' == import_path[0]:  # 如果是相对路径。
                # import_path = import_path[1:]  # 相对路径会多一个点。
                raise RuntimeError(f"Relative import is not supported yet. ")
            module = import_module(import_path)
        except Exception:
            # print(traceback.format_exc())
            warnings.warn(f"未能用import_module导入{py},尝试直接执行文件。")

    if module is None:
        spec = spec_from_file_location("foo", py)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)

    return module



#!/usr/bin/env src
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: networks.py
@time: 2020/3/23 22:29
@desc:
"""
from typing import Dict, Tuple, List, Union

import torch
from torch import nn

from alchemy_cat.torch_tools.network.update_model_state_dict import strip_named_params, \
    update_model_state_dict, find_match_miss, kParamDictType

__all__ = ["Network"]


class Network(nn.Module):
    """Torch base networks with default I/O info and initialization"""

    in_batch_size = None
    out_batch_size = None
    in_channel = None
    out_channel = None
    in_height = None
    out_height = None
    in_width = None
    out_width = None

    def initialize(self, pretrain_dict: kParamDictType, verbosity: int=2, **kwargs):
        """Default initialize. Initialize every sub module with type Network using pretrain_dict

        Args:
            pretrain_dict: pretrain state dict
            verbosity: 0: No info; 1: Show weight loaded; 2: Show weight missed; 3: Show weight Redundant. (Default: 2)
        """
        for child in self.children():
            if isinstance(child, Network):
                child.initialize(pretrain_dict, verbosity, **kwargs)

    def find_match_miss(self, pretrain_dict: kParamDictType) \
            -> Tuple[Union[Dict[str, nn.Parameter], Dict[str, torch.Tensor]], List[str]]:
        """Find params in pretrain dict with the same suffix name of params of model.

        Args:
            pretrain_dict: pretrain dict with key: parameter

        Returns:
            matched_dict: keys with corresponding pretrain values of model's state_dict whose suffix name can be found
                in pretrain_dict
            missed_keys: keys of model's state_dict whose suffix name can not be found in pretrain_dict
        """
        return find_match_miss(self, pretrain_dict)

    def update_model_state_dict(self, weight_dict: kParamDictType, verbosity: int=2):
        """Update model's state_dict with pretrain_dict

        Args:
            weight_dict: pretrain state dict
            verbosity: 0: No info; 1: Show weight loaded; 2: Show weight missed; 3: Show weight Redundant
        """
        update_model_state_dict(self, weight_dict, verbosity=verbosity)

    @property
    def strip_named_params(self):
        return strip_named_params(self.state_dict())



#!/usr/bin/env src
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: update_model_state_dict.py
@time: 2020/3/23 5:42
@desc:
"""
from typing import Dict, Tuple, List, Union

import re
from colorama import Style, Fore
import torch

import torch.nn as nn

__all__ = ['update_model_state_dict', 'strip_param_name', 'strip_named_params', 'named_param_match_pattern',
           'find_match_miss', 'kParamDictType', 'sub_module_name_of_named_params']

kParamDictType = Union[Dict[str, nn.Parameter], Dict[str, torch.Tensor]]


def update_model_state_dict(model: nn.Module, weight_dict: kParamDictType, verbosity: int = 2):
    """Update model's state_dict with pretrain_dict

        Args:
            model: model to be updated
            weight_dict: pretrain state dict
            verbosity: 0: No info; 1: Show weight loaded; 2: Show weight missed; 3: Show weight Redundant
    """
    model_dict = model.state_dict()
    update_dict = {k: weight_dict[k] for k in model_dict.keys() if k in weight_dict}
    model_dict.update(update_dict)
    model.load_state_dict(model_dict, strict=True)

    if verbosity >= 1:
        for k in sorted(update_dict.keys()):
            print(Fore.GREEN + "[WEIGHT LOADED] " + Style.RESET_ALL
                  + "{:<50s} shape: {}".format(k, update_dict[k].shape))

    if verbosity >= 2:
        missing_dict = {k: model_dict[k] for k in model_dict.keys() if k not in weight_dict}
        for k in sorted(missing_dict.keys()):
            print(Fore.YELLOW + "[WEIGHT MISSED] " + Style.RESET_ALL +
                  "{:<50s} shape: {}".format(k, missing_dict[k].shape))

    if verbosity >= 3:
        redundant_dict = {k: weight_dict[k] for k in weight_dict.keys() if k not in model_dict}
        for k in sorted(redundant_dict.keys()):
            print(Fore.LIGHTCYAN_EX + "[WEIGHT UNUSED] " + Style.RESET_ALL +
                  "{:<50s} shape: {}".format(k, redundant_dict[k].shape))


def strip_param_name(param_name: str) -> str:
    """Input an module's param name, return it's origin name with out parent modules' name

    Args:
        param_name (str): parameter name with it's parent module prefix

    Returns:
        Param's origin name
    """
    splits = param_name.rsplit('.', maxsplit=2)
    return '.'.join(splits[-2:])


def strip_named_params(named_params: kParamDictType) -> Union[Dict[str, nn.Parameter], Dict[str, torch.Tensor]]:
    """Strip all param names in modules' state_dict

    Args:
        named_params: module's state_dict

    Returns:
        A Dict with key of stripped param name and it's origin value.

    See Also:
        strip_param_name
    """
    return {strip_param_name(k): v for k, v in named_params.items()}


def named_param_match_pattern(pattern: str, named_param: kParamDictType,
                              full_match: bool = False, multi_match=True) -> kParamDictType:
    """Find key-value pair in named_param whose key match pattern

    Args:
        pattern: re pattern
        named_param: named parameters
        full_match: If True, the key matched pattern must exactly equal to matched string. (Default: False)
        multi_match: If False, when multi keys match pattern, raise Error. (Default: True)

    Returns:
        Key-value pair whose key match pattern

    Raises:
        RuntimeError:
            More than one key matched
    """
    matched_key_value_pairs = {}
    for key in named_param.keys():
        re_ret = re.search(pattern, key)
        if re_ret is not None:
            if full_match and re_ret.group(0) != key:
                continue
            else:
                matched_key_value_pairs[key] = named_param[key]

    if len(matched_key_value_pairs) > 1 and not multi_match:
        raise RuntimeError(f"More than one parameter {matched_key_value_pairs} can by matched by pattern {pattern}")

    return matched_key_value_pairs


def find_match_miss(model: nn.Module, pretrain_dict: kParamDictType) \
        -> Tuple[Union[Dict[str, nn.Parameter], Dict[str, torch.Tensor]], List[str]]:
    """Find params in pretrain dict with the same suffix name of params of model.

        Args:
            model: torch model
            pretrain_dict: pretrain dict with key: parameter

        Returns:
            matched_dict: keys with corresponding pretrain values of model's state_dict whose suffix name can be found
                in pretrain_dict
            missed_keys: keys of model's state_dict whose suffix name can not be found in pretrain_dict
        """
    matched_dict = dict()
    missed_keys = []
    strip_pretrain_dict = strip_named_params(pretrain_dict)

    for model_param_name in model.state_dict().keys():
        strip_model_param_name = strip_param_name(model_param_name)
        if strip_model_param_name in strip_pretrain_dict:
            matched_dict[model_param_name] = strip_pretrain_dict[strip_model_param_name]
        else:
            missed_keys.append(model_param_name)

    return matched_dict, missed_keys


def sub_module_name_of_named_params(named_params: kParamDictType, module_name_sub_dict: Dict[str, str]) \
    -> Union[Dict[str, nn.Parameter], Dict[str, torch.Tensor]]:
    """Sub named_parameters key's module name part with module_name_sub_dict.

    Args:
        named_params: Key-value pair of param name and param value.
        module_name_sub_dict: Module names' sub dict.

    Returns:
        named parameters whose module name part of it's param name is subbed by module_name_sub_dict.
    """
    sub_named_params = dict()
    for module_param_name, value in named_params.items():
        param_name, module_name = map(lambda inverse_name: inverse_name[::-1],
                                      module_param_name[::-1].split('.', maxsplit=1))
        if module_name not in module_name_sub_dict:
            sub_named_params[module_param_name] = value
        else:
            sub_named_params[module_name_sub_dict[module_name] + '.' + param_name] = value
    return sub_named_params



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/5/22 14:12
@File    : stat_metrics.py
@Software: PyCharm
@Desc    : 
"""
from math import sqrt
from functools import partial

import xlwings as xw
import numpy as np
import pandas as pd
import scipy.stats as st


@xw.func(async_mode='threading')
@xw.arg('X', np.ndarray, dtype=np.float64, doc='所有样本。')
@xw.ret(numbers=int)
def 样本数(X: np.ndarray) -> int:
    return X.size


@xw.func(async_mode='threading')
@xw.arg('X', np.ndarray, dtype=np.float64, doc='所有样本。')
@xw.arg('ddof', numbers=int, doc='归一化时，自由度的减数。')
@xw.ret(numbers=float)
def 标准差(X: np.ndarray, ddof: int = 1) -> float:
    return X.std(ddof=ddof)


@xw.func(async_mode='threading')
@xw.arg('X', np.ndarray, dtype=np.float64, doc='所有样本。')
@xw.ret(numbers=float)
def 均值(X: np.ndarray) -> float:
    return X.mean()


@xw.func(async_mode='threading')
@xw.arg('X', np.ndarray, doc='所有样本。')
@xw.ret(transpose=True)
def 最小最大值(X: np.ndarray) -> np.ndarray:
    return np.array([np.amin(X), np.amax(X)])


@xw.func(async_mode='threading')
@xw.arg('alpha', numbers=float, doc='1 - 置信度。')
@xw.arg('X', np.ndarray, dtype=np.float64, doc='所有样本。')
@xw.arg('sigma', numbers=float, doc='样本所来自分布的标准差，若缺省，则用t分布计算置信区间。')
@xw.ret(transpose=True)
def 置信区间(alpha: float, X: np.ndarray, sigma: float | None = None) -> np.ndarray:
    if sigma is None:
        return np.array(st.t.interval(alpha=1 - alpha, df=X.size - 1,
                                      loc=X.mean(), scale=st.sem(X, axis=None)), dtype=np.float64)
    else:
        return np.array(st.norm.interval(alpha=1 - alpha, loc=X.mean(),
                                         scale=sigma / (X.size ** 0.5)), dtype=np.float64)


@xw.func(async_mode='threading')
@xw.arg('alpha', numbers=float, doc='1 - 置信度。')
@xw.arg('X', np.ndarray, dtype=np.float64, doc='所有样本。')
@xw.arg('sigma', numbers=float, doc='样本所来自分布的标准差，若缺省，则用t分布计算置信区间。')
@xw.ret(transpose=True)
def 单侧置信区间(alpha: float, X: np.ndarray, sigma: float | None = None) -> np.ndarray:
    return 置信区间(alpha * 2, X, sigma)


@xw.func(async_mode='threading')
@xw.arg('metrics', pd.DataFrame, dtype=np.float64, index=0, header=1, doc='所有样本的各种指标。')
@xw.arg('alpha', numbers=float, doc='1 - 置信度。')
@xw.arg('sigma', numbers=float, ndim=1, doc='样本所来自分布的标准差，若缺省，则用t分布计算置信区间。')
@xw.ret(index=True, header=True)
def 总结多次实验(metrics: pd.DataFrame, alpha: float=.05, sigma: list | None = None) -> pd.DataFrame:
    if sigma is not None:
        assert len(sigma) == len(metrics.columns)
    else:
        sigma = [sigma] * len(metrics.columns)

    index = ['样本数', '均值', '无偏标准差', '有偏标准差', '最小值', '最大值',
             '置信区间下限', '置信区间上限', '置信区间半径', '单侧置信下限', '单侧置信上限']
    conclude = pd.DataFrame(data=np.full((len(index), len(metrics.columns)), np.nan, dtype=np.float64),
                            index=index, columns=metrics.columns)

    for column, s in zip(conclude.columns, sigma):
        con, X = conclude[column], metrics[column].values
        con['样本数'] = 样本数(X)
        con['均值'] = 均值(X)
        con['无偏标准差'] = 标准差(X, ddof=1)
        con['有偏标准差'] = 标准差(X, ddof=0)
        con['最小值'], con['最大值'] = 最小最大值(X)
        con['置信区间下限'], con['置信区间上限'] = 置信区间(alpha, X, s)
        con['置信区间半径'] = (con['置信区间上限'] - con['置信区间下限']) / 2
        con['单侧置信下限'], con['单侧置信上限'] = 单侧置信区间(alpha, X, s)

    return conclude


@xw.func(async_mode='threading')
@xw.arg('alpha', numbers=float, doc='1 - 置信度。')
@xw.arg('n_X', numbers=int, doc='总体X的样本总数。')
@xw.arg('n_Y', numbers=int, doc='总体Y的样本总数。')
@xw.arg('avg_X', numbers=float, doc='总体X的样本均值。')
@xw.arg('avg_Y', numbers=float, doc='总体Y的样本均值。')
@xw.arg('sigma_X', numbers=float, doc='总体X所来自分布的标准差，或样本的标准差。')
@xw.arg('sigma_Y', numbers=float, doc='总体Y所来自分布的标准差，或样本的标准差。')
@xw.arg('use_t', numbers=bool, doc='是否使用t分布计算置信区间，默认为True。')
@xw.ret(transpose=True)
def 总体间均值差置信区间(alpha: float,
                         n_X: int, n_Y: int,
                         avg_X: float, avg_Y: float,
                         sigma_X: float, sigma_Y: float,
                         use_t: bool = True) -> np.ndarray:
    loc = avg_X - avg_Y

    if not use_t:
        scale = sqrt((sigma_X ** 2 / n_X) + (sigma_Y ** 2 / n_Y))
    else:
        df = n_X + n_Y - 2
        scale = sqrt((((n_X - 1) * sigma_X ** 2) + ((n_Y - 1) * sigma_Y ** 2)) / df)
        scale *= sqrt((1 / n_X) + (1 / n_Y))

    if not use_t:
        return np.array(st.norm.interval(alpha=1 - alpha, loc=loc, scale=scale), dtype=np.float64)
    else:
        return np.array(st.t.interval(alpha=1 - alpha, df=df, loc=loc, scale=scale), dtype=np.float64)


@xw.func(async_mode='threading')
@xw.arg('alpha', numbers=float, doc='1 - 置信度。')
@xw.arg('n_X', numbers=int, doc='总体X的样本总数。')
@xw.arg('n_Y', numbers=int, doc='总体Y的样本总数。')
@xw.arg('avg_X', numbers=float, doc='总体X的样本均值。')
@xw.arg('avg_Y', numbers=float, doc='总体Y的样本均值。')
@xw.arg('sigma_X', numbers=float, doc='总体X所来自分布的标准差，或样本的标准差。')
@xw.arg('sigma_Y', numbers=float, doc='总体Y所来自分布的标准差，或样本的标准差。')
@xw.arg('use_t', numbers=bool, doc='是否使用t分布计算置信区间，默认为True。')
@xw.ret(transpose=True)
def 总体间均值差单侧置信区间(alpha: float,
                             n_X: int, n_Y: int,
                             avg_X: float, avg_Y: float,
                             sigma_X: float, sigma_Y: float,
                             use_t: bool = True) -> np.ndarray:
    return 总体间均值差置信区间(alpha * 2, n_X, n_Y, avg_X, avg_Y, sigma_X, sigma_Y, use_t)


def _单个总体检验统计量(hypo_X: float, n_X: int, avg_X: float, sigma_X: float) -> dict[str, float]:
    ret = {'loc': hypo_X, 'scale': sigma_X / sqrt(n_X)}
    ret['statistics'] = (avg_X - hypo_X) / ret['scale']
    return ret


@xw.func(async_mode='threading')
@xw.arg('hypo_X', numbers=float, doc='假设均值。')
@xw.arg('n_X', numbers=int, doc='总体X的样本总数。')
@xw.arg('avg_X', numbers=float, doc='总体X的样本均值。')
@xw.arg('sigma_X', numbers=float, doc='总体X所来自分布的标准差，或样本的标准差。')
@xw.arg('alternative', doc='备择假设；gt：样本均值大于假设，lt：样本均值小于假设，'
                           'eq：样本均值等于假设，ned：样本均值不等于假设。')
@xw.arg('use_t', numbers=bool, doc='是否使用t分布计算P值，默认为True。')
def 单个总体P值(hypo_X: float,
                n_X: int, avg_X: float, sigma_X: float,
                alternative: str | int,
                use_t: bool = True) -> float:
    统计量 = _单个总体检验统计量(hypo_X, n_X, avg_X, sigma_X)
    loc, scale, statistics = 统计量['loc'], 统计量['scale'], 统计量['statistics']
    sf = partial(st.t.sf, df=n_X - 1) if use_t else st.norm.sf
    cdf = partial(st.t.cdf, df=n_X - 1) if use_t else st.norm.cdf
    match alternative:
        case 'gt' | 0:
            p_val = sf(statistics)
        case 'lt' | 1:
            p_val = cdf(statistics)
        case 'eq' | 2:
            p_val = 1 - 2 * sf(abs(statistics))
        case 'neq' | 3:
            p_val = 2 * sf(abs(statistics))
        case _:
            raise ValueError(f"不支持的{alternative=}。")
    return p_val


def _两个总体检验统计量(hypo_X_sub_Y: float,
                        n_X: int, n_Y: int,
                        avg_X: float, avg_Y: float,
                        sigma_X: float, sigma_Y: float,
                        use_t: bool = True
                        ) -> dict[str, float]:
    ret = {'loc': hypo_X_sub_Y}
    if not use_t:
        ret['scale'] = sqrt((sigma_X ** 2 / n_X) + (sigma_Y ** 2 / n_Y))
    else:
        df = n_X + n_Y - 2
        ret['scale'] = sqrt((((n_X - 1) * sigma_X ** 2) + ((n_Y - 1) * sigma_Y ** 2)) / df)
        ret['scale'] *= sqrt((1 / n_X) + (1 / n_Y))
    ret['statistics'] = (avg_X - avg_Y - hypo_X_sub_Y) / ret['scale']
    return ret


xw.func(async_mode='threading')


@xw.arg('hypo_X_sub_Y', numbers=float, doc='假设均值之差。')
@xw.arg('n_X', numbers=int, doc='总体X的样本总数。')
@xw.arg('n_Y', numbers=int, doc='总体Y的样本总数。')
@xw.arg('avg_X', numbers=float, doc='总体X的样本均值。')
@xw.arg('avg_Y', numbers=float, doc='总体Y的样本均值。')
@xw.arg('sigma_X', numbers=float, doc='总体X所来自分布的标准差，或样本的标准差。')
@xw.arg('sigma_Y', numbers=float, doc='总体Y所来自分布的标准差，或样本的标准差。')
@xw.arg('alternative', doc='备择假设；gt：样本均值差大于假设，lt：样本均值差小于假设，'
                           'eq：样本均值差等于假设，ned：样本均值差不等于假设。')
@xw.arg('use_t', numbers=bool, doc='是否使用t分布计算P值，默认为True。')
def 两个总体P值(hypo_X_sub_Y: float,
                n_X: int, n_Y: int,
                avg_X: float, avg_Y: float,
                sigma_X: float, sigma_Y: float,
                alternative: str | int,
                use_t: bool = True) -> float:
    统计量 = _两个总体检验统计量(hypo_X_sub_Y,
                                 n_X, n_Y,
                                 avg_X, avg_Y,
                                 sigma_X, sigma_Y,
                                 use_t=use_t)
    loc, scale, statistics = 统计量['loc'], 统计量['scale'], 统计量['statistics']
    sf = partial(st.t.sf, df=n_X + n_Y - 2) if use_t else st.norm.sf
    cdf = partial(st.t.cdf, df=n_X + n_Y - 2) if use_t else st.norm.cdf
    match alternative:
        case 'gt' | 0:
            p_val = sf(statistics)
        case 'lt' | 1:
            p_val = cdf(statistics)
        case 'eq' | 2:
            p_val = 1 - 2 * sf(abs(statistics))
        case 'neq' | 3:
            p_val = 2 * sf(abs(statistics))
        case _:
            raise ValueError(f"不支持的{alternative=}。")
    return p_val


xw.func(async_mode='threading')


@xw.arg('avg_X', numbers=float, doc='总体X的样本均值。')
@xw.arg('radius_X', numbers=float, doc='样本均值置信区间半径。')
@xw.arg('mul', numbers=float, doc='显示时乘以因子。')
@xw.arg('ndigits', numbers=int, doc='保留几位小数。')
def 格式化均值加减区间半径(avg_X: float, radius_X: float, mul: float = 100., ndigits: int = 2) -> str:
    return f'{avg_X * mul:.{ndigits}f}±{radius_X * mul:.{ndigits}f}'


if __name__ == '__main__':
    xw.serve()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2024/7/10 4:16
@File    : test_yacs_yaml.py
@Software: PyCharm
@Desc    : 
"""
import sys
from pathlib import Path

import pytest

from alchemy_cat.py_tools import Config, load_config, open_config, Cfg2Tune, Param2Tune

YAML_CFG = 'py_tools/config/tests/migrate/yaml.yaml'
YACS_CFG = 'py_tools/config/tests/migrate/yacs_cfg.py'
YACS_GET_CFG = 'py_tools/config/tests/migrate/yacs_get.py'
MM_CFG = 'py_tools/config/tests/migrate/mm_configs/deeplabv3plus/deeplabv3plus_r50-d8_4xb2-40k_cityscapes-512x1024.py'
AC_CFG = 'py_tools/config/tests/migrate/ac_cfg.py'

sys.path = ['', 'py_tools/config/tests'] + sys.path


def is_yaml() -> bool:
    try:
        import yaml
        return True
    except ImportError:
        return False


def is_yacs() -> bool:
    try:
        from yacs.config import CfgNode
        return True
    except ImportError:
        return False


def is_mm() -> bool:
    try:
        from mmengine import Config
        return True
    except ImportError:
        return False


@pytest.fixture(scope='module')
def dump_py():
    dump_py = Path('/tmp/test_alchemy_cat_config/dump.py')
    dump_py.parent.mkdir(parents=True, exist_ok=True)
    yield dump_py
    # dump_py.unlink(missing_ok=True)


@pytest.fixture(scope='module')
def dump_yaml():
    dump_yaml = Path('/tmp/test_alchemy_cat_config/dump.yaml')
    dump_yaml.parent.mkdir(parents=True, exist_ok=True)
    yield dump_yaml
    # dump_yaml.unlink(missing_ok=True)


@pytest.fixture(scope='function')
def mm_config2tune():
    cfg = Cfg2Tune(caps=MM_CFG)
    cfg.rslt_dir = 'xxx'
    cfg.model.backbone.depth = Param2Tune([50, 101])
    cfg.train_cfg.max_iters = Param2Tune([10_000, 20_000])
    yield cfg


@pytest.mark.skipif(not is_yaml(), reason="yaml is not installed")
def test_open_yaml():
    import yaml

    loaded_cfg = load_config(YAML_CFG, create_rslt_dir=False).unfreeze()
    del loaded_cfg.rslt_dir

    init_cfg = Config(YAML_CFG)

    caps_cfg = Config(caps=YAML_CFG)
    caps_cfg.rslt_dir = 'xxx'
    caps_cfg.parse(create_rslt_dir=False).unfreeze()
    del caps_cfg.rslt_dir

    with open(YAML_CFG, 'r') as f:
        yaml_cfg = yaml.load(f, Loader=yaml.FullLoader)

    assert yaml_cfg == loaded_cfg == init_cfg == caps_cfg


@pytest.mark.skipif(not is_yacs(), reason="yacs is not installed")
def test_open_yacs_cfg():
    loaded_cfg = load_config(YACS_CFG, create_rslt_dir=False).unfreeze()
    del loaded_cfg.rslt_dir

    init_cfg = Config(YACS_CFG)

    caps_cfg = Config(caps=YACS_CFG)
    caps_cfg.rslt_dir = 'xxx'
    caps_cfg.parse(create_rslt_dir=False).unfreeze()
    del caps_cfg.rslt_dir

    from migrate.yacs_cfg import cfg as yacs_cfg

    assert yacs_cfg == loaded_cfg == init_cfg == caps_cfg


@pytest.mark.skipif(not is_yacs(), reason="yacs is not installed")
def test_open_yacs_get_cfg():
    loaded_cfg = load_config(YACS_GET_CFG, create_rslt_dir=False).unfreeze()
    del loaded_cfg.rslt_dir

    init_cfg = Config(YACS_GET_CFG)

    caps_cfg = Config(caps=YACS_GET_CFG)
    caps_cfg.rslt_dir = 'xxx'
    caps_cfg.parse(create_rslt_dir=False).unfreeze()
    del caps_cfg.rslt_dir

    from migrate.yacs_get import get_cfg_defaults

    assert get_cfg_defaults() == loaded_cfg == init_cfg == caps_cfg


@pytest.mark.skipif(not is_mm(), reason="mmengine is not installed")
def test_open_mm():
    loaded_cfg = load_config(MM_CFG, create_rslt_dir=False).unfreeze()
    del loaded_cfg.rslt_dir

    init_cfg = Config(MM_CFG)

    caps_cfg = Config(caps=MM_CFG)
    caps_cfg.rslt_dir = 'xxx'
    caps_cfg.parse(create_rslt_dir=False).unfreeze()
    del caps_cfg.rslt_dir

    from mmengine import Config as MMConfig
    mm_cfg = MMConfig.fromfile(MM_CFG)._cfg_dict

    assert mm_cfg == loaded_cfg == init_cfg == caps_cfg


def test_to_cfg(dump_py: Path):
    x_cfg = open_config(AC_CFG)[0]

    Config.from_x_to_y(AC_CFG, dump_py, y_type='alchemy-cat')

    y_cfg = open_config(str(dump_py))[0]

    assert x_cfg == y_cfg


@pytest.mark.skipif(not is_yaml(), reason="yaml is not installed")
def test_to_yaml(dump_yaml: Path):
    x_cfg = open_config(AC_CFG)[0]

    Config.from_x_to_y(AC_CFG, dump_yaml, y_type='yaml')

    y_cfg = open_config(str(dump_yaml))[0]

    assert x_cfg == y_cfg


@pytest.mark.skipif(not is_mm(), reason="mmengine is not installed")
def test_to_mm(dump_py: Path):
    x_cfg = open_config(AC_CFG)[0]

    Config.from_x_to_y(AC_CFG, dump_py, y_type='mmcv')

    y_cfg = open_config(str(dump_py))[0]

    assert x_cfg == y_cfg


@pytest.mark.skipif(not is_mm(), reason="mmengine is not installed")
def test_tune_mm(mm_config2tune: Cfg2Tune, dump_py: Path):
    from mmengine import Config as MMConfig

    mm_cfg = MMConfig.fromfile(MM_CFG)

    cfgs = list(mm_config2tune.get_cfgs())

    cfgs[0].save_mmcv(dump_py)
    mm_cfg0 = MMConfig.fromfile(str(dump_py))
    del mm_cfg0._cfg_dict['rslt_dir']
    assert mm_cfg0.model.backbone.depth == 50
    assert mm_cfg0.train_cfg.max_iters == 10_000
    mm_cfg0.model.backbone.depth = 50
    mm_cfg0.train_cfg.max_iters = 40_000
    assert mm_cfg0._cfg_dict == mm_cfg._cfg_dict

    cfgs[1].save_mmcv(dump_py)
    mm_cfg1 = MMConfig.fromfile(str(dump_py))
    del mm_cfg1._cfg_dict['rslt_dir']
    assert mm_cfg1.model.backbone.depth == 50
    assert mm_cfg1.train_cfg.max_iters == 20_000
    mm_cfg1.model.backbone.depth = 50
    mm_cfg1.train_cfg.max_iters = 40_000
    assert mm_cfg1._cfg_dict == mm_cfg._cfg_dict

    cfgs[2].save_mmcv(dump_py)
    mm_cfg2 = MMConfig.fromfile(str(dump_py))
    del mm_cfg2._cfg_dict['rslt_dir']
    assert mm_cfg2.model.backbone.depth == 101
    assert mm_cfg2.train_cfg.max_iters == 10_000
    mm_cfg2.model.backbone.depth = 50
    mm_cfg2.train_cfg.max_iters = 40_000
    assert mm_cfg2._cfg_dict == mm_cfg._cfg_dict

    cfgs[3].save_mmcv(dump_py)
    mm_cfg3 = MMConfig.fromfile(str(dump_py))
    del mm_cfg3._cfg_dict['rslt_dir']
    assert mm_cfg3.model.backbone.depth == 101
    assert mm_cfg3.train_cfg.max_iters == 20_000
    mm_cfg3.model.backbone.depth = 50
    mm_cfg3.train_cfg.max_iters = 40_000
    assert mm_cfg3._cfg_dict == mm_cfg._cfg_dict



# -- configs/patches/cifar10.py --

import torchvision.transforms as T
from torchvision.datasets import CIFAR10

from alchemy_cat.dl_config import Config

cfg = Config()

cfg.dt.set_whole()
cfg.dt.cls = CIFAR10
cfg.dt.ini.root = '/tmp/data'
cfg.dt.ini.transform = T.Compose([T.ToTensor(), T.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])



""" Bigger example on solar PV modeling using PVLib

https://github.com/pvlib/pvlib-python

Example: https://github.com/pvlib/pvlib-python/blob/master/docs/tutorials/tmy_to_power.ipynb
"""

import inspect
import os
from pprint import pprint
import warnings
try:
    import pvlib
except ImportError:
    raise ImportError('This example require pvlib')
from pvlib import pvsystem
from alchemy_cat.dag.core import Graph


warnings.filterwarnings("ignore")


pvlib_abspath = os.path.dirname(os.path.abspath(inspect.getfile(pvlib)))
datapath = os.path.join(pvlib_abspath, 'data', '703165TY.csv')
tmy_data, meta = pvlib.tmy.readtmy3(datapath, coerce_year=2015)
tmy_data.index.name = 'Time'
tmy_data = tmy_data.shift(freq='-30Min')
sandia_modules = pvlib.pvsystem.retrieve_sam(name='SandiaMod')
cec_modules = pvlib.pvsystem.retrieve_sam(name='CECMod')
sapm_inverters = pvlib.pvsystem.retrieve_sam('sandiainverter')
sapm_inverter = sapm_inverters['ABB__MICRO_0_25_I_OUTD_US_208_208V__CEC_2014_']

data = {
    'index': tmy_data.index,
    'surface_tilt': 30,
    'surface_azimuth': 180,
    'DHI': tmy_data['DHI'],
    'DNI': tmy_data['DNI'],
    'GHI': tmy_data['GHI'],
    'Wspd': tmy_data['Wspd'],
    'DryBulb': tmy_data['DryBulb'],
    'albedo': 0.2,
    'latitude': meta['latitude'],
    'longitude': meta['longitude'],
    'sandia_module': sandia_modules.Canadian_Solar_CS5P_220M___2009_,
    'cec_module': cec_modules.Canadian_Solar_CS5P_220M,
    'alpha_sc': cec_modules.Canadian_Solar_CS5P_220M['alpha_sc'],
    'EgRef': 1.121,
    'dEgdT': -0.0002677,
    'inverter': sapm_inverter
}


def unpack_df(f):
    def wrap(*args, **kwargs):
        res = f(*args, **kwargs)
        return res.values.T
    return wrap


def unpack_dict(f):  # (ordered)
    def wrap(*args, **kwargs):
        res = f(*args, **kwargs)
        return [res[k] for k in res]
    return wrap


# parallelism not needed for this example
graph = Graph(parallel=False)

graph.add_node(
    unpack_df(pvlib.solarposition.get_solarposition),
    inputs=['index', 'latitude', 'longitude'],
    outputs=['apparent_elevation', 'apparent_zenith', 'azimuth',
             'elevation', 'equation_of_time', 'zenith']
)
graph.add_node(
    pvlib.irradiance.extraradiation,
    inputs=['index'],
    outputs=['dni_extra']
)
graph.add_node(
    pvlib.atmosphere.relativeairmass,
    inputs=['apparent_zenith'],
    outputs=['airmass']
)
graph.add_node(
    pvlib.irradiance.haydavies,
    inputs=['surface_tilt', 'surface_azimuth', 'DHI', 'DNI',
            'dni_extra', 'apparent_zenith', 'azimuth'],
    outputs=['poa_sky_diffuse']
)
graph.add_node(
    pvlib.irradiance.grounddiffuse,
    inputs=['surface_tilt', 'GHI', 'albedo'],
    outputs=['poa_ground_diffuse']
)
graph.add_node(
    pvlib.irradiance.aoi,
    inputs=['surface_tilt', 'surface_azimuth', 'apparent_zenith', 'azimuth'],
    outputs=['aoi']
)
graph.add_node(
    unpack_df(pvlib.irradiance.globalinplane),
    inputs=['aoi', 'DNI', 'poa_sky_diffuse', 'poa_ground_diffuse'],
    outputs=['poa_global', 'poa_direct', 'poa_diffuse']
)
graph.add_node(
    unpack_df(pvlib.pvsystem.sapm_celltemp),
    inputs=['poa_global', 'Wspd', 'DryBulb'],
    outputs=['temp_cell', 'temp_module']
)
graph.add_node(
    pvlib.pvsystem.sapm_effective_irradiance,
    inputs=['poa_direct', 'poa_diffuse', 'airmass', 'aoi', 'sandia_module'],
    outputs=['effective_irradiance'],
)
graph.add_node(
    unpack_dict(pvlib.pvsystem.sapm),
    inputs=['effective_irradiance', 'temp_cell', 'sandia_module'],
    outputs=['i_sc_sapm', 'i_mp_sapm', 'v_oc_sapm', 'v_mp_sapm', 'p_mp_sapm', 'i_x_sapm', 'i_xx_sapm']
)
graph.add_node(
    pvlib.pvsystem.calcparams_desoto,
    inputs=['poa_global', 'temp_cell', 'alpha_sc', 'cec_module', 'EgRef', 'dEgdT'],
    outputs=['photocurrent', 'saturation_current', 'resistance_series',
             'resistance_shunt', 'nNsVth']
)
graph.add_node(
    unpack_dict(pvlib.pvsystem.singlediode),
    inputs=['photocurrent', 'saturation_current', 'resistance_series', 'resistance_shunt', 'nNsVth'],
    outputs=['i_sc_sd', 'i_mp_sd', 'v_oc_sd', 'v_mp_sd', 'p_mp_sd', 'i_x_sd', 'i_xx_sd']
)
graph.add_node(
    pvlib.pvsystem.snlinverter,
    inputs=['v_mp_sapm', 'p_mp_sapm', 'inverter'],
    outputs=['pac_sapm']
)
graph.add_node(
    pvlib.pvsystem.snlinverter,
    inputs=['v_mp_sd', 'p_mp_sd', 'inverter'],
    outputs=['pac_sd']
)

pprint(graph.dag)
graph.calculate(data)
print(sum(graph.data['pac_sd']))
print(sum(graph.data['pac_sapm']))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/11/26 22:25
@File    : to_cuda.py
@Software: PyCharm
@Desc    : 
"""
import torch

__all__ = ['dict_to_cuda']


def dict_to_cuda(inp: dict):
    for k, v in inp.items():
        if torch.is_tensor(v):
            inp[k] = v.to('cuda', non_blocking=True)
        if isinstance(v, dict):
            dict_to_cuda(v)
    return inp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2024/7/8 15:08
@File    : __init__.py.py
@Software: PyCharm
@Desc    : py_tools绝不知道dl_config的存在，dl_config cherry-pick出py_tools的配置工具，并提供example。
           外人建议通过dl_config来访问配置工具。
"""
from ..py_tools.config import *
from ..py_tools.param_tuner import *



# -- configs/tune/tune_bs_epoch,subject_to/cfg.py --

from alchemy_cat.dl_config import Cfg2Tune, Param2Tune

cfg = Cfg2Tune(caps='configs/mnist/base,sched_from_addon/cfg.py')

cfg.loader.ini.batch_size = Param2Tune([128, 256, 512])

cfg.sched.epochs = Param2Tune([5, 15],
                              subject_to=lambda cur_val: cur_val * cfg.loader.ini.batch_size.cur_val <= 15 * 128)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/8/2 21:09
@File    : py3_server.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np

from alchemy_cat.contrib.shm_send_recv.py3_send_recv import server_send, server_recv, init_server  # 导入server所需函数

kImageSize = (512, 1024)

kClientObj = list(np.arange(100))

kServerObj = {'a': 1, (3,): 2.2, 'c': 'str', 'd': [1, 2]}
kServerArr = np.ones(kImageSize, dtype=np.float32) * 255
kServerPredict = np.ones(kImageSize, dtype=np.uint8) * 255


init_server()  # 初始化server，非必须。可以防止多个server同时运行，且保证server被中断后（譬如被kill）后，buffer能被正确清理。

recv_count = 0
while True:
    img = server_recv()  # 接收图片
    assert img.shape == kImageSize + (3,)
    assert np.all(img == 1)

    arr_recv = server_recv()  # 接收numpy array
    assert arr_recv.shape == kImageSize
    assert np.all(arr_recv == 1.)
    assert arr_recv.dtype == np.float32

    obj_recv = server_recv()  # 接收python对象
    assert obj_recv == kClientObj

    server_send(kServerObj)  # 发送python对象
    server_send(kServerArr)  # 发送numpy array
    server_send(kServerPredict)  # 发送预测结果

    recv_count += 1
    print("Server has received for %d imgs" % recv_count)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/16 18:11
@File    : label2cls_in_label.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np

from alchemy_cat.contrib.voc import VOC_CLASSES, VOC

__all__ = ['label2cls_in_label']


def label2cls_in_label(label: np.ndarray, cls_num=len(VOC_CLASSES), ignore_label=VOC.ignore_label) -> np.ndarray:
    """将像素级标签转为图像级标签。默认按照VOC配置。

    Args:
        label: 像素级标签。
        cls_num: 类别数。
        ignore_label: “忽略”类别的编号。

    Returns:
        图像级标签。
    """
    return (np.bincount(label.ravel(), minlength=ignore_label + 1) != 0).astype(np.int64)[:cls_num]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/4/13 21:15
@File    : module_init_fns.py
@Software: PyCharm
@Desc    : 
"""
from torch import nn

__all__ = ['init_conv_kaiming_bn_identity', 'set_bn_momentum']


def init_conv_kaiming_bn_identity(module: nn.Module) -> nn.Module:
    for m in module.modules():
        if isinstance(m, (nn.modules.conv._ConvNd, nn.Linear)):  # _ConvNd下辖1、2、3D正反卷积，拥有weight和bias两个参数。
            nn.init.kaiming_normal_(m.weight, mode='fan_in', nonlinearity='relu')
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.modules.batchnorm._NormBase):
            # _NormBase下辖1、2、3D及同步批归一化，拥有weight和bias两个参数，running mean、var等buffer。
            nn.init.constant_(m.weight, 1)
            nn.init.constant_(m.bias, 0)

    return module


def set_bn_momentum(model: nn.Module, momentum: float=0.1):
    """设置网络BN层的动量。

    Args:
        model: torch模型。
        momentum: 动量。
    """
    for m in model.modules():
        if isinstance(m, nn.modules.batchnorm._BatchNorm):
            m.momentum = momentum



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/7 11:36
@File    : manip_ch.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np
import torch

from typing import Union

__all__ = ["HWC2CHW", "CHW2HWC", "BGR2RGB", "RGB2BGR", "c1_to_cn"]


def HWC2CHW(arr: np.ndarray) -> np.ndarray:
    """
    Args:
        arr (np.ndarray): ndarray with shape (..., H, W, C)

    Returns:
        (np.ndarray): ndarray transposed to (..., C, H, W)
    """
    return arr.transpose(tuple(range(0, arr.ndim - 3)) + (-1, -3, -2))


def CHW2HWC(arr: np.ndarray) -> np.ndarray:
    """
    Args:
        arr (np.ndarray): ndarray with shape (..., C, H, W)

    Returns:
        (np.ndarray): ndarray transposed to (..., H, W, C)
    """
    return arr.transpose(tuple(range(0, arr.ndim - 3)) + (-2, -1, -3))


def RGB2BGR(arr: Union[torch.Tensor, np.ndarray]) -> Union[torch.Tensor, np.ndarray]:
    """
    Args:
        arr (Tensor, ndarray): arr with shape (..., C[RGB])

    Returns:
        arr with shape (..., C[BGR])
    """
    return arr[..., ::-1]


def BGR2RGB(arr: Union[torch.Tensor, np.ndarray]) -> Union[torch.Tensor, np.ndarray]:
    """
    Args:
        arr (Tensor, ndarray): arr with shape (..., C[BGR])

    Returns:
        arr with shape (..., C[RGB])
    """
    return arr[..., ::-1]


def c1_to_cn(c1_array: np.ndarray, n: int=3):
    """Convert 1 channel img/imgs to n channel imgs by repeat in last dimension

    Args:
        c1_array: ndarray img/imgs
        n: Channel num of converted img

    Returns:
        Converted img with shape (c1_array.shape, n)
    """
    cn_array = np.zeros(c1_array.shape + (n, ), dtype=c1_array.dtype)
    for i in range(n):
        cn_array[..., i] = c1_array
    return cn_array



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/16 14:35
@File    : resize_cams.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np
import torch
import torch.nn.functional as F
from alchemy_cat.alg import size2HW
import cv2

__all__ = ['resize_cam', 'resize_cam_cuda']


def resize_cam(cam: np.ndarray, size, interpolation: int=cv2.INTER_LINEAR) -> np.ndarray:
    """将cam采样到指定尺寸。

    Args:
        cam: (N, ori_h, ori_w)的cam。
        size: (h, w)的尺寸，若为int，则为H，W相同。
        interpolation: 插值方式。

    Returns:
        (N, H, W)的cam。
    """
    ori_h, ori_w = cam.shape[1:]
    h, w = size2HW(size)

    if (ori_h, ori_w) != (h, w):
        cam = cv2.resize(cam.transpose(1, 2, 0), (w, h), interpolation=interpolation)
        if cam.ndim == 2:  # N=1时，cv2.resize会丢掉维度。
            cam = cam[None, :, :]
        else:
            cam = cam.transpose(2, 0, 1)

    return cam


def resize_cam_cuda(cam: torch.Tensor, size, interpolation: str='bilinear') -> torch.Tensor:
    h, w = size2HW(size)
    need_resize = (cam.shape[1] != h) or (cam.shape[2] != w)

    return F.interpolate(cam.unsqueeze(0), size=(h, w), mode=interpolation,
                         align_corners=False).squeeze(0) if need_resize else cam



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/2/22 20:19
@File    : voc上色.py
@Software: PyCharm
@Desc    : 
"""
import argparse
import os
import os.path as osp

from alchemy_cat.contrib.voc import VOCAug2
from alchemy_cat.data.plugins import arr2PIL
from tqdm import tqdm

parser = argparse.ArgumentParser()
parser.add_argument('-t', '--target', type=str)
parser.add_argument('-s', '--split', type=str)
parser.add_argument('-a', '--alpha', type=float, default=0.5)
args = parser.parse_args()

os.makedirs(args.target, exist_ok=True)

dt = VOCAug2('datasets', split=args.split)

for inp in tqdm(dt, dynamic_ncols=True, desc="处理", unit="张"):
    img_id, img, lb = inp['img_id'], inp['img'], inp['lb']
    color_lb = dt.label_map2color_map(lb)

    img_lb = (img * args.alpha + color_lb * (1 - args.alpha)).astype('uint8')

    arr2PIL(img_lb, order='RGB').save(osp.join(args.target, f'{img_id}.png'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: test_data_manager.py
@time: 2020/3/21 13:31
@desc:
"""
import pytest

import numpy as np
import sys
import torch

from alchemy_cat.contrib.voc import VOCAug, VOCTrainAuger
from alchemy_cat.data import DataManager, Subset
from alchemy_cat.py_tools.random import set_rand_seed
from alchemy_cat.data.tests.voc2012_auger_with_sub_graph import VOCTrainAuger as VOCTrainAugerSubGraph


sub_voc_aug = Subset(VOCAug(split='val'), list(range(50)))
sub_voc_aug.mean_bgr = VOCAug.mean_bgr
sub_voc_aug.ignore_label = VOCAug.ignore_label

voc_auger = VOCTrainAuger(sub_voc_aug, slim=True,
                          scale_factors=(0.5, 1.0), is_multi_mirror=True, is_color_jitter=True)
voc_auger_sub_graph = VOCTrainAugerSubGraph(sub_voc_aug, slim=True,
                                            scale_factors=(0.5, 1.0), is_multi_mirror=True, is_color_jitter=True)

if sys.platform == 'win32':
    voc_data_manager = DataManager(dataset=sub_voc_aug, data_auger=voc_auger, log_dir='./Temp/test_data_manager',
                                   batch_size=10, shuffle=True)
else:
    voc_data_manager = DataManager(dataset=sub_voc_aug, data_auger=voc_auger, log_dir='./Temp/test_data_manager',
                                   batch_size=10, shuffle=True, num_workers=4)

if sys.platform == 'win32':
    voc_data_manager_sub_graph = DataManager(dataset=sub_voc_aug, data_auger=voc_auger_sub_graph,
                                             log_dir='./Temp/test_data_manager/test_sub_graph',
                                             batch_size=10, shuffle=True)
else:
    voc_data_manager_sub_graph = DataManager(dataset=sub_voc_aug, data_auger=voc_auger_sub_graph,
                                             log_dir='./Temp/test_data_manager/test_sub_graph',
                                             batch_size=10, shuffle=True, num_workers=4)


def setup_function(func):
    set_rand_seed(0)
    print(f"-----------setup function: set_rand_seed(0)-----------")


@pytest.fixture(scope='module')
def data_manager():
    return voc_data_manager


@pytest.fixture(scope='function', params=[voc_data_manager, voc_data_manager_sub_graph], ids=['no_sub', 'with_sub'])
def data_manager_with_sub_graph(request):
    return request.param


def test_recover_and_recreate(data_manager_with_sub_graph):
    """Test Recover and Recreate

        * Run 15 iter with 'start_epoch', record their imgs
        * Move to iter0. Run 30 iter with 'next_batch', test whether the first 150 iter have the same img. Record imgs.
        * Restart at iter15 with 'start_epoch', run 15 iter. Test whether they are the same to recorded img.
        * Move to iter15 with and run 15 iter in recreate mode with 'next_batch'. Test whether the first 5 iter are
          the same to the record img while the last not.

    Args:
        data_manager: voc data manager with 1 epoch == 10 iteration
    """
    record_imgs = []

    epoch_iter = data_manager_with_sub_graph.start_epoch()
    for i in range(15):
        try:
            batch = next(epoch_iter)
        except StopIteration:
            epoch_iter = data_manager_with_sub_graph.start_epoch()
            batch = next(epoch_iter)
        img_ids, imgs, labels = batch
        imgs = imgs.numpy()

        record_imgs.append(imgs)

    data_manager_with_sub_graph.move_epoch_to(0, 0)
    for i in range(30):
        batch = data_manager_with_sub_graph.next_batch()
        img_ids, imgs, labels = batch
        imgs = imgs.numpy()

        if i < 15:
            assert np.all(record_imgs[i] == imgs)
        else:
            record_imgs.append(imgs)

    epoch_iter = data_manager_with_sub_graph.start_epoch(15)
    for i in range(15, 30):
        try:
            batch = next(epoch_iter)
        except StopIteration:
            epoch_iter = data_manager_with_sub_graph.start_epoch()
            batch = next(epoch_iter)
        img_ids, imgs, labels = batch
        imgs = imgs.numpy()

        assert np.all(record_imgs[i] == imgs)

    data_manager_with_sub_graph.move_epoch_to(1, 5)
    for i in range(15, 30):
        batch = data_manager_with_sub_graph.next_batch(recreate=True)
        img_ids, imgs, labels = batch
        imgs = imgs.numpy()

        if i < 20:
            assert np.all(record_imgs[i] == imgs)
        else:
            assert np.any(record_imgs[i] != imgs)


def test_batches2indices(data_manager):
    data_manager.move_epoch_to(0, 0)
    batches = data_manager.epoch_batches
    indices = data_manager.batches2indices(batches)

    flatten_batches = np.array(batches).ravel().tolist()
    assert indices == flatten_batches
    assert data_manager.epoch_indices == indices
    indices.sort()
    assert indices == list(range(100))


def test_batch_num(data_manager):
    assert len(data_manager) == 10
    assert data_manager.batch_num == 10


def test_backtrace(data_manager):
    ret = data_manager.backtrace(0, epoch_loc=0)

    data_manager.move_epoch_to(0, 0)
    batch = data_manager.next_batch(recreate=True)
    img_ids, imgs, labels = batch
    img = imgs.numpy()[0]

    assert np.all(ret['auger_output'][1] == img)
    print(ret)


@pytest.mark.skipif('sys.platform == "win32"')
def test_worker_init_fn():
    def worker_ini_fn(id):
        print(f"WorkerID: {id} with torch seed: {torch.initial_seed()}")

    data_manager = DataManager(dataset=sub_voc_aug, data_auger=voc_auger, log_dir='./Temp/test_data_manager',
                                   batch_size=10, shuffle=True, num_workers=4, worker_init_fn=worker_ini_fn)

    epoch_iter = data_manager.start_epoch()
    for i in range(15):
        try:
            batch = next(epoch_iter)
        except StopIteration:
            epoch_iter = data_manager.start_epoch()
            batch = next(epoch_iter)
